from nawah.base_module import BaseModule
from nawah.enums import Event
from nawah.classes import ATTR, PERM, EXTN, METHOD
from nawah.config import Config
from nawah.registry import Registry

from bson import ObjectId


class Realm(BaseModule):
	'''`Realm` module module provides data type and controller for Realm Mode in Nawah eco-system.'''

	collection = 'realms'
	attrs = {
		'user': ATTR.ID(
			desc='`_id` of `User` doc the doc belongs to. This is also the ADMIN of the realm.'
		),
		'name': ATTR.STR(
			desc='Name of the realm. This is both readable as well as unique name.'
		),
		'default': ATTR.ID(
			desc='`_id` of `Group` doc that serves as `DEFAULT` group of the realm.'
		),
		'create_time': ATTR.DATETIME(
			desc='Python `datetime` ISO format of the doc creation.'
		),
	}
	methods = {
		'read': METHOD(permissions=[PERM(privilege='read')]),
		'create': METHOD(permissions=[PERM(privilege='create')]),
		'update': METHOD(
			permissions=[
				PERM(privilege='update', doc_mod={'user': None, 'create_time': None})
			],
			query_args={'_id': ATTR.ID()},
		),
		'delete': METHOD(
			permissions=[PERM(privilege='delete')],
			query_args={'_id': ATTR.ID()},
		),
	}

	async def pre_create(self, skip_events, env, query, doc, payload):
		user_doc = {attr: doc['user'][attr] for attr in Config.user_attrs}
		user_doc.update(
			{
				'locale': Config.locale,
				'groups': [],
				'privileges': {'*': '*'},
				'status': 'active',
				'attrs': {},
				'realm': doc['name'],
			}
		)
		user_results = await Registry.module('user').create(
			skip_events=[Event.PERM, Event.ARGS, Event.PRE], env=env, doc=user_doc
		)
		if user_results.status != 200:
			return user_results
		user = user_results.args.docs[0]

		group_results = await Registry.module('group').create(
			skip_events=[Event.PERM, Event.ARGS],
			env=env,
			doc={
				'user': user._id,
				'name': {locale: '__DEFAULT' for locale in Config.locales},
				'bio': {locale: '__DEFAULT' for locale in Config.locales},
				'privileges': Config.default_privileges,
				'attrs': {},
				'realm': doc['name'],
			},
		)
		if group_results.status != 200:
			return group_results
		group = group_results.args.docs[0]

		skip_events.append(Event.ARGS)
		doc['user'] = user._id
		doc['default'] = group._id
		return (skip_events, env, query, doc, payload)

	async def on_create(self, results, skip_events, env, query, doc, payload):
		for doc in results['docs']:
			realm_results = await self.read(
				skip_events=[Event.PERM, Event.ARGS], env=env, query=[{'_id': doc._id}]
			)
			realm = realm_results.args.docs[0]
			Config._realms[realm.name] = realm
			Config._sys_docs[realm._id] = {'module': 'realm'}
		return (results, skip_events, env, query, doc, payload)



from nawah.config import Config
from nawah.base_module import BaseModule
from nawah.classes import SYS_DOC

from typing import Dict, List, Any

import logging

logger = logging.getLogger('nawah')


class InvalidModuleException(Exception):
	def __init__(self, *, module):
		self.module = module
		logger.debug(f'InvalidModuleException: {str(self)}')

	def __str__(self):
		return f'Invalid module \'{self.module}\''


class InvalidLocaleException(Exception):
	def __init__(self, *, locale):
		self.locale = locale
		logger.debug(f'InvalidLocaleException: {str(self)}')

	def __str__(self):
		return f'Invalid locale \'{self.locale}\''


class InvalidLocaleTermException(Exception):
	def __init__(self, *, locale, term):
		self.locale = locale
		self.term = term
		logger.debug(f'InvalidLocaleTermException: {str(self)}')

	def __str__(self):
		return f'Invalid term \'{self.term}\' of locale \'{self.locale}\''


class InvalidVarException(Exception):
	def __init__(self, *, var):
		self.var = var
		logger.debug(f'InvalidVarException: {str(self)}')

	def __str__(self):
		return f'Invalid var \'{self.var}\''


class Registry:
	docs: List[SYS_DOC] = Config.docs
	jobs: List[Dict[str, Any]] = Config.jobs

	@staticmethod
	def module(module: str) -> BaseModule:
		try:
			return Config.modules[module]
		except KeyError:
			raise InvalidModuleException(module=module)

	@staticmethod
	def l10n(*, locale: str, term: str):
		if locale not in Config.l10n.keys():
			raise InvalidLocaleException(locale=locale)

		try:
			if '.' in term:
				from nawah.utils import extract_attr

				return extract_attr(scope=Config.l10n[locale], attr_path=term)
			else:
				return Config.l10n[locale][term]
		except (KeyError, TypeError):
			raise InvalidLocaleTermException(locale=locale, term=term)

	@staticmethod
	def var(var: str):
		try:
			if '.' in var:
				from nawah.utils import extract_attr

				return extract_attr(scope=Config.vars, attr_path=var)
			else:
				return Config.vars[var]
		except (KeyError, TypeError):
			raise InvalidVarException(var=var)


from ._generate_attr import generate_attr
from ._attr import (
	deep_update,
	extract_attr,
	set_attr,
	update_attr_values,
	expand_attr,
)
from ._validate import (
	InvalidAttrException,
	MissingAttrException,
	ConvertAttrException,
	validate_doc,
	validate_attr,
	process_file_obj,
	generate_dynamic_attr,
)
from ._import_modules import import_modules
from ._generate_ref import generate_ref, extract_lambda_body
from ._generate_models import generate_models
from ._encode_attr_type import encode_attr_type


from nawah.config import Config
from nawah.enums import Event
from nawah.classes import (
	ATTR,
	JSONEncoder,
	Query,
	NAWAH_QUERY,
	NAWAH_DOC,
	NAWAH_ENV,
	DictObj,
)
from nawah.utils import extract_attr, validate_attr, generate_attr

from bson import ObjectId
from typing import List, Dict, Union, Tuple, Literal, Any, TypedDict, cast, Iterable

import logging, traceback, datetime, os, json, copy

logger = logging.getLogger('test')
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s [TEST] [%(levelname)s]  %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.setLevel(logging.DEBUG)


RESULTS_STATS = TypedDict(
	'RESULTS_STATS', {'passed': int, 'failed': int, 'skipped': int, 'total': int}
)
RESULTS = TypedDict(
	'RESULTS',
	{
		'test': List['STEP'],
		'status': Literal['PASSED', 'PARTIAL', 'FAILED', True, False],
		'success_rate': int,
		'stats': RESULTS_STATS,
		'steps': List[Dict[str, Any]],
	},
)


class InvalidTestStepException(Exception):
	def __init__(self, *, msg: str):
		self.msg = msg

	def __str__(self):
		return self.msg


class TEST(list):
	pass


class STEP:
	_step: Literal['AUTH', 'SIGNOUT', 'CALL', 'TEST', 'ASSIGN_VAR', 'SET_REALM']
	_args: Dict[str, Any]

	def __init__(
		self,
		*,
		step: Literal['AUTH', 'SIGNOUT', 'CALL', 'TEST', 'ASSIGN_VAR', 'SET_REALM'],
		**kwargs: Any,
	):
		self._step = step
		self._args = kwargs

	@classmethod
	def AUTH(cls, *, var: str, val: str, hash: str):
		return STEP(step='AUTH', var=var, val=val, hash=hash)

	@classmethod
	def SIGNOUT(cls):
		return STEP(step='SIGNOUT')

	@classmethod
	def CALL(
		cls,
		*,
		module: str,
		method: str,
		skip_events: List[str] = None,
		query: NAWAH_QUERY = None,
		doc: NAWAH_DOC = None,
		acceptance: Dict[str, Any] = {'status': 200},
	):
		if not skip_events:
			skip_events = []
		if not query:
			query = []
		if not doc:
			doc = {}
		return STEP(
			step='CALL',
			module=module,
			method=method,
			skip_events=skip_events,
			query=query,
			doc=doc,
			acceptance=acceptance,
		)

	@classmethod
	def TEST(cls, *, test: str, steps: Union[List[int], range] = None):
		return STEP(step='TEST', test=test, steps=steps)

	@classmethod
	def ASSIGN_VAR(cls, *, vars: dict):
		return STEP(step='ASSIGN_VAR', vars=vars)

	@classmethod
	def SET_REALM(cls, *, realm: str):
		return STEP(step='SET_REALM')

	@classmethod
	async def validate_step(cls, *, step: 'STEP'):
		logger.debug(f'Attempting to validate test step: {step}')
		if step._step == 'AUTH':
			logger.debug(f'Validating test step \'AUTH\' with args: {step._args}')
			if step._args['var'] not in Config.user_attrs.keys():
				raise InvalidTestStepException(
					msg=f'Test step arg \'var\' is invalid. Found \'{step._args["var"]}\', but expecting one of: {list(Config.user_attrs.keys())}.'
				)
			if step._args['val'].startswith('$__'):
				logger.debug(
					'Detected test step \'AUTH\' with Test Variable \'val\' arg. Skipping validating test arg \'val\'.'
				)
			else:
				try:
					await validate_attr(
						mode='create',
						attr_name='val',
						attr_type=Config.user_attrs[step._args['var']],
						attr_val=step._args['val'],
					)
				except InvalidTestStepException:
					raise InvalidTestStepException(
						msg=f'Test step arg \'val\' of type \'{type(step._args["val"])}\' is invalid with required type \'{Config.user_attrs[step._args["var"]]._type}\'.'
					)
		elif step._step == 'SIGNOUT':
			logger.debug('Skipping validating test step \'SIGNOUT\'.')
		elif step._step == 'CALL':
			logger.debug(f'Validating test step \'CALL\' with args: {step._args}')
			if step._args['module'] not in Config.modules.keys():
				raise InvalidTestStepException(
					msg=f'Test step arg \'module\' is invalid. Unknown Nawah module \'{step._args["module"]}\'.'
				)
			if step._args['method'] not in Config.modules[step._args['module']].methods.keys():
				raise InvalidTestStepException(
					msg=f'Test step arg \'method\' is invalid. Unknown method \'{step._args["method"]}\'.'
				)
		elif step._step == 'TEST':
			logger.debug(f'Validating test step \'TEST\' with args: {step._args}')
			if step._args['test'] not in Config.tests.keys():
				raise InvalidTestStepException(
					msg=f'Test step arg \'test\' is invalid. Unknown test \'{step._args["test"]}\'.'
				)
			if step._args['steps'] != None and (
				type(step._args['steps']) not in [list, range]
				or sum(step for step in step._args['steps'] if type(step) != int)
			):
				raise InvalidTestStepException(
					msg='Test step arg \'steps\' is invalid. Either value is not of type \'list\', nor \'range\' or at least on item in the list is not of type \'int\'.'
				)
		elif step._step == 'ASSIGN_VAR':
			logger.debug('Skipping validating test step \'ASSIGN_VAR\'.')
		else:
			raise InvalidTestStepException(
				msg=f'Unknown test step \'{step._step}\' with args: {step._args}.'
			)


class CALC:
	_opers: Dict[str, str] = {
		'+': '__add__',
		'-': '__sub__',
		'*': '__mul__',
		'/': '__truediv__',
		'**': '__pow__',
	}
	_oper: Literal['+', '-', '*', '/', '**']
	_attrs: List[Any]

	def __init__(self, *, oper: Literal['+', '-', '*', '/', '**'], attrs: List[Any]):
		self._oper = oper
		self._attrs = attrs

	def execute(self, *, scope: Dict[str, Any]):
		results = None
		for i in range(len(self._attrs)):
			# [DOC] Attempt to extract attr, execute oper
			if type(self._attrs[i]) == str and self._attrs[i].startswith('$__'):
				self._attrs[i] = extract_attr(scope=scope, attr_path=self._attrs[i])
			elif type(self._attrs[i]) in [CALC, CAST, JOIN]:
				self._attrs[i] = self._attrs[i].execute(scope=scope)
			# [DOC] Calculate results per _oper
			if i == 0:
				results = self._attrs[i]
			else:
				results = getattr((results if results else 0), self._opers[self._oper])(
					self._attrs[i]
				)
		return results


class CAST:
	_type: Literal['int', 'float', 'str']
	_attr: Any

	def __init__(self, *, type: Literal['int', 'float', 'str'], attr: Any):
		self._type = type
		self._attr = attr

	def execute(self, *, scope: Dict[str, Any]):
		# [DOC] Attempt to extract attr, execute oper
		if type(self._attr) == str and self._attr.startswith('$__'):
			self._attr = extract_attr(scope=scope, attr_path=self._attr)
		elif type(self._attr) in [CALC, CAST, JOIN]:
			self._attr = self._attr.execute(scope=scope)
		# [DOC] Cast per _type
		if self._type == 'int':
			return int(self._attr)
		elif self._type == 'float':
			return float(self._attr)
		elif self._type == 'str':
			return str(self._attr)


class JOIN:
	_separator: str
	_attrs: List[Any]

	def __init__(self, *, separator: str, attrs: List[Any]):
		self._separator = separator
		self._attrs = attrs

	def execute(self, *, scope: Dict[str, Any]):
		for i in range(len(self._attrs)):
			# [DOC] Attempt to extract attr, execute oper
			if type(self._attrs[i]) == str and self._attrs[i].startswith('$__'):
				self._attrs[i] = extract_attr(scope=scope, attr_path=self._attrs[i])
			elif type(self._attrs[i]) in [CALC, CAST, JOIN]:
				self._attrs[i] = self._attrs[i].execute(scope=scope)
		# [DOC] Join using _separator
		return self._separator.join(self._attrs)


class Test:
	session: DictObj
	env: NAWAH_ENV
	vars: Dict[str, Any] = {}

	@classmethod
	async def run_test(cls, test_name: str, steps: List[STEP] = None) -> RESULTS:
		if test_name not in Config.tests.keys():
			logger.error('Specified test is not defined in loaded config.')
			logger.debug(f'Loaded tests: {list(Config.tests.keys())}')
			exit(1)
		test: List[STEP] = Config.tests[test_name]
		results: RESULTS = {
			'test': Config.tests[test_name],
			'status': 'PASSED',
			'success_rate': 100,
			'stats': {'passed': 0, 'failed': 0, 'skipped': 0, 'total': 0},
			'steps': [],
		}

		step_failed = False
		for i in range(len(test)):
			results['stats']['total'] += 1
			step = copy.deepcopy(test[i])

			if steps and i not in steps:
				results['stats']['total'] -= 1
				results['stats']['skipped'] += 1
				continue

			if step_failed and not Config.test_force:
				results['stats']['skipped'] += 1
				continue

			if step._step == 'AUTH':
				try:
					await STEP.validate_step(step=step)
					step = STEP.CALL(
						module='session',
						method='auth',
						doc={
							step._args['var']: step._args['val'],
							'hash': step._args['hash'],
						},
					)
				except InvalidTestStepException as e:
					logger.error(f'Can\'t process test step \'AUTH\' with error: {e} Exiting.')
					exit(1)
			elif step._step == 'SIGNOUT':
				step = STEP.CALL(module='session', method='signout', query=[{'_id': '$__session'}])
			else:
				try:
					await STEP.validate_step(step=step)
				except InvalidTestStepException as e:
					logger.error(f'{e} Exiting.')
					exit(1)

			if step._step == 'CALL':
				call_results = await cls.run_call(results=results, **step._args)
				if 'session' in call_results.keys():
					logger.debug('Updating session after detecting \'session\' in call results.')
					if str(call_results['session']._id) == 'f00000000000000000000012':
						cls.session = DictObj(
							{
								**Config.compile_anon_session(),
								'user': DictObj(Config.compile_anon_user()),
							}
						)
					else:
						cls.session = call_results['session']
				results['steps'].append(call_results)

			elif step._step == 'TEST':
				test_results = await cls.run_test(
					test_name=step._args['test'], steps=step._args['steps']
				)
				if test_results['status'] == 'PASSED':
					test_results['status'] = True
				else:
					test_results['status'] = False
				results['steps'].append(test_results)  # type: ignore

			elif step._step == 'ASSIGN_VAR':
				vars = {}
				for var in step._args['vars'].keys():
					vars[var] = extract_attr(scope=results, attr_path=step._args['vars'][var])  # type: ignore
					cls.vars[var] = vars[var]
				results['steps'].append({'step': 'assign_var', 'vars': vars, 'status': True})

			elif step._step == 'SET_REALM':
				try:
					if step._args['realm'].startswith('$__'):
						step._args['realm'] = extract_attr(scope=results, attr_path=step._args['realm'])  # type: ignore
					logger.debug(f'Updating realm to \'{step._args["realm"]}\'.')
					cls.env['realm'] = step._args['realm']
					results['steps'].append(
						{
							'step': 'set_realm',
							'realm': step._args['realm'],
							'status': True,
						}
					)
				except Exception as e:
					logger.error(e)
					results['steps'].append(
						{
							'step': 'set_realm',
							'realm': step._args['realm'],
							'status': False,
						}
					)

			if not results['steps'][-1]['status']:
				results['stats']['failed'] += 1
				if not Config.test_force:
					step_failed = True
			else:
				results['stats']['passed'] += 1

		if len(results['steps']) == 0:
			logger.error('No steps tested. Exiting.')
			exit(1)

		results['success_rate'] = int(
			(results['stats']['passed'] / results['stats']['total']) * 100
		)
		if results['success_rate'] == 0:
			results['status'] = 'FAILED'
		elif results['success_rate'] == 100:
			results['status'] = 'PASSED'
		else:
			results['status'] = 'PARTIAL'

		if test_name == Config.test_name:
			logger.debug(
				'Finished testing %s steps [Passed: %s, Failed: %s, Skipped: %s] with success rate of: %s%%',
				results['stats']['total'],
				results['stats']['passed'],
				results['stats']['failed'],
				results['stats']['skipped'],
				results['success_rate'],
			)
			tests_log = os.path.join(
				Config._app_path,
				'tests',
				f'NAWAH-TEST_{test_name}_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}',
			)
			if os.path.exists(f'{tests_log}.json'):
				i = 1
				while True:
					if os.path.exists(f'{tests_log}.{i}.json'):
						i += 1
					else:
						tests_log = f'{tests_log}.{i}'
						break
			tests_log += '.json'
			with open(tests_log, 'w') as f:
				f.write(json.dumps(json.loads(JSONEncoder().encode(results)), indent=4))
				logger.debug(f'Full tests log available at: {tests_log}')
			if results['success_rate'] == 100:
				exit(0)
			else:
				exit(1)
		else:
			return results

	@classmethod
	async def run_call(
		cls,
		results: RESULTS,
		module: str,
		method: str,
		skip_events: List[Event],
		query: List[Any],
		doc: NAWAH_DOC,
		acceptance: Dict[str, Any],
	):
		# [DOC] If query, doc of call are callable call them
		if callable(query):
			query = query(results=results)
		if callable(doc):
			doc = doc(results=results)

		call_results: Dict[str, Any] = {
			'step': 'call',
			'module': module,
			'method': method,
			'query': query,
			'doc': doc,
			'status': True,
		}
		query = Query(cls.process_obj(results=results, obj=query))  # type: ignore
		doc = cls.process_obj(results=results, obj=doc)  # type: ignore
		try:
			call_results['results'] = await Config.modules[module].methods[method](
				skip_events=skip_events,
				env={**cls.env, 'session': cls.session},
				query=query,
				doc=doc,
				call_id='__TEST__',
			)
			call_results['acceptance'] = cls.process_obj(
				results=results,
				obj=copy.deepcopy(acceptance),
				call_results=call_results,
			)
			for measure in acceptance.keys():
				if measure.startswith('session.'):
					results_measure = extract_attr(
						scope=cls.session,
						attr_path=f'$__{measure.replace("session.", "", 1)}',
					)
				else:
					results_measure = extract_attr(
						scope=call_results['results'], attr_path=f'$__{measure}'
					)
				if results_measure != call_results['acceptance'][measure]:
					call_results['status'] = False
					cls.break_debugger(locals(), call_results)
					break
			if call_results['status'] == False:
				logger.debug(
					f'Test step \'call\' failed at measure \'{measure}\'. Required value is \'{call_results["acceptance"][measure]}\', but test results is \'{results_measure}\''
				)
				call_results['measure'] = measure
		except Exception as e:
			tb = traceback.format_exc()
			logger.error(f'Exception occurred: {tb}')
			cls.break_debugger(locals(), call_results)
			call_results.update(
				{
					'measure': measure,
					'results': {
						'status': 500,
						'msg': str(e),
						'args': {'tb': tb, 'code': 'SERVER_ERROR'},
					},
				}
			)
			call_results['status'] = False
			call_results['measure'] = measure
		if call_results['status'] == True and 'session' in call_results['results'].args:
			call_results['session'] = call_results['results'].args.session
		return call_results

	@classmethod
	def process_obj(
		cls,
		results: RESULTS,
		obj: Union[Dict[str, Any], List[Any]],
		call_results: Union[Dict[str, Any], 'DictObj'] = None,
	) -> Union[Dict[str, Any], List[Any]]:
		logger.debug(f'Attempting to process object: {obj}')
		obj_iter: Iterable
		if type(obj) == dict:
			obj = cast(Dict[str, Any], obj)
			obj_iter = obj.keys()
		elif type(obj) == list:
			obj = cast(List[Any], obj)
			obj_iter = range(len(obj))
		else:
			logger.error(
				f'Object is not of types \'dict\' or \'list\'. Refer to log to check how invalid type obj was attempted to be processed. Exiting.'
			)
			exit(1)

		for j in obj_iter:
			if type(obj[j]) == ATTR:
				# [DOC] Check for impossible to generate KV_DICT Attr Type value where min+len(req) > max
				if obj[j]._type == 'KV_DICT':
					if obj[j]._args['max']:
						if (obj[j]._args['min'] or 0) + len(obj[j]._args['req'] or []) > obj[j]._args[
							'max'
						]:
							logger.error(
								f'Attr Generator of type \'KV_DICT\' where Attr Arg \'min\' and length of Attr Arg \'req\' are greater than Attr Arg \'max\' can\'t be generated. Exiting.'
							)
							exit(1)
				obj[j] = generate_attr(attr_type=obj[j])
			elif type(obj[j]) in [CALC, CAST, JOIN]:
				obj[j] = obj[j].execute(scope=results)
			elif type(obj[j]) == dict:
				obj[j] = cls.process_obj(results=results, obj=obj[j], call_results=call_results)
			elif type(obj[j]) == list:
				if len(obj[j]) and type(obj[j][0]) == dict and '__attr' in obj[j][0].keys():
					if 'count' not in obj[j][0].keys():
						obj[j][0]['count'] = 1
					obj[j] = [
						generate_attr(attr_type=obj[j][0]['__attr']) for ii in range(obj[j][0]['count'])
					]
				else:
					obj[j] = cls.process_obj(results=results, obj=obj[j], call_results=call_results)
			elif type(obj[j]) == str and obj[j].startswith('$__'):
				if obj[j] == '$__session':
					obj[j] = cls.session
				elif obj[j].startswith('$__session.'):
					obj[j] = extract_attr(
						scope=cls.session, attr_path=obj[j].replace('session.', '', 1)
					)
				elif obj[j].startswith('$__vars.'):
					obj[j] = extract_attr(scope=cls.vars, attr_path=obj[j].replace('vars.', '', 1))
				else:
					obj[j] = extract_attr(scope=results, attr_path=obj[j])  # type: ignore
			elif callable(obj[j]):
				obj[j] = obj[j](results=results, call_results=call_results)

		logger.debug(f'Processed object: {obj}')
		return obj

	@classmethod
	def break_debugger(cls, scope: Dict[str, Any], call_results: Dict[str, Any]) -> None:
		if Config.test_breakpoint:
			logger.debug(
				'Creating a breakpoint to allow you to investigate step failure. Type \'c\' after finishing to continue.'
			)
			logger.debug('All variables are available under \'scope\' dict.')
			if call_results:
				logger.debug('Call test raised exception available under \'call_results\' dict')
			breakpoint()



from nawah.enums import Event

from typing import (
	Tuple,
	Dict,
	Any,
	Optional,
	Union,
	List,
	TYPE_CHECKING,
	AsyncGenerator,
	Callable,
	Literal,
	Protocol,
	TypedDict,
)

import datetime

if TYPE_CHECKING:
	from nawah.base_method import BaseMethod
	from ._attr import ATTR
	from ._dictobj import DictObj
	from ._types import NAWAH_EVENTS, NAWAH_ENV, NAWAH_DOC, NAWAH_QUERY
	from ._query import Query


PRE_HANDLER_RETURN = Tuple[
	'NAWAH_EVENTS', 'NAWAH_ENV', 'Query', 'NAWAH_DOC', Dict[str, Any]
]
ON_HANDLER_RETURN = Tuple[
	Dict[str, Any], 'NAWAH_EVENTS', 'NAWAH_ENV', 'Query', 'NAWAH_DOC', Dict[str, Any]
]

PERM_QUERY_MOD_UNIT = Dict[str, Union['ATTR_MOD', Literal['$__date', '$__user'], Any]]
PERM_QUERY_MOD = Union[PERM_QUERY_MOD_UNIT, 'NAWAH_QUERY', List['PERM_QUERY_MOD']]  # type: ignore

PERM_DOC_MOD_UNIT = Dict[str, Union['ATTR_MOD', Literal['$__date', '$__user'], Any]]
PERM_DOC_MOD = Union[PERM_DOC_MOD_UNIT, 'NAWAH_DOC']


class MethodException(Exception):
	pass


class ATTR_MOD_CONDITION(Protocol):
	def __call__(
		self,
		skip_events: 'NAWAH_EVENTS',
		env: 'NAWAH_ENV',
		query: Union['Query', 'NAWAH_QUERY'],
		doc: 'NAWAH_DOC',
		scope: Optional['NAWAH_DOC'],
	) -> bool:
		...


class ATTR_MOD_DEFAULT(Protocol):
	def __call__(
		self,
		skip_events: 'NAWAH_EVENTS',
		env: 'NAWAH_ENV',
		query: Union['Query', 'NAWAH_QUERY'],
		doc: 'NAWAH_DOC',
		scope: Optional['NAWAH_DOC'],
	) -> Any:
		...


class ATTR_MOD:
	condition: ATTR_MOD_CONDITION
	default: Union[ATTR_MOD_DEFAULT, Any]

	def __repr__(self):
		return f'<ATTR_MOD:{self.condition},{self.default}>'

	def __init__(
		self,
		*,
		condition: ATTR_MOD_CONDITION,
		default: Union[ATTR_MOD_DEFAULT, Any],
	):
		setattr(self, 'condition', condition)
		self.default = default


class PERM:
	privilege: str
	query_mod: PERM_QUERY_MOD
	doc_mod: PERM_DOC_MOD

	def __repr__(self):
		return f'<PERM:{self.privilege},{self.query_mod},{self.doc_mod}>'

	def __init__(
		self,
		*,
		privilege: str,
		query_mod: Optional[PERM_QUERY_MOD] = None,
		doc_mod: Optional[PERM_DOC_MOD] = None,
	):
		if not query_mod:
			query_mod = {}
		if not doc_mod:
			doc_mod = {}
		self.privilege = privilege
		self.query_mod = query_mod
		self.doc_mod = doc_mod


class EXTN:
	module: str
	skip_events: Optional[List[Event]]
	query: Optional['NAWAH_QUERY']
	attrs: Union[List[str], str]
	force: Union[bool, str] = False

	def __repr__(self):
		return f'<EXTN:{self.module},{self.attrs},{self.force}>'

	def __init__(
		self,
		*,
		module: str,
		skip_events: List[Event] = None,
		query: 'NAWAH_QUERY' = None,
		attrs: Union[List[str], str],
		force: Union[bool, str] = False,
	):
		self.module = module
		self.skip_events = skip_events
		self.query = query
		self.attrs = attrs
		self.force = force

		# [DOC] Wrap query in list if it is a dict
		if type(query) == dict:
			self.query = [self.query]


class METHOD:
	permissions: List[PERM]
	query_args: Union[
		None,
		List[Dict[str, 'ATTR']],
		Dict[str, 'ATTR'],
	]
	doc_args: Union[
		None,
		List[Dict[str, 'ATTR']],
		Dict[str, 'ATTR'],
	] = None
	get_method: bool
	post_method: bool
	watch_method: bool
	_callable: 'BaseMethod'

	def __init__(
		self,
		*,
		permissions: List[PERM],
		query_args: Union[
			None,
			List[Dict[str, 'ATTR']],
			Dict[str, 'ATTR'],
		] = None,
		doc_args: Union[
			None,
			List[Dict[str, 'ATTR']],
			Dict[str, 'ATTR'],
		] = None,
		get_method: bool = False,
		post_method: bool = False,
		watch_method: bool = False,
	):
		self.permissions = permissions
		self.query_args = query_args
		self.doc_args = doc_args
		self.get_method = get_method
		self.post_method = post_method
		self.watch_method = watch_method

	def __call__(self, **kwargs):
		return self._callable(**kwargs)


class CACHE_CONDITION(Protocol):
	def __call__(
		self,
		skip_events: 'NAWAH_EVENTS',
		env: 'NAWAH_ENV',
		query: Union['Query', 'NAWAH_QUERY'],
	) -> bool:
		...


class CACHE:
	condition: CACHE_CONDITION
	period: Optional[int]
	queries: Dict[str, 'CACHED_QUERY']

	def __repr__(self):
		return f'<CACHE:{self.condition},{self.period}>'

	def __init__(
		self,
		*,
		condition: CACHE_CONDITION,
		period: int = None,
	):
		setattr(self, 'condition', condition)
		self.period = period
		self.queries = {}

	def cache_query(self, *, query_key: str, results: Dict[str, Any]):
		self.queries[query_key] = CACHED_QUERY(results=results)


class CACHED_QUERY:
	results: Dict[str, Any]
	query_time: datetime.datetime

	def __init__(self, *, results: Dict[str, Any], query_time: datetime.datetime = None):
		self.results = results
		if not query_time:
			query_time = datetime.datetime.utcnow()
		self.query_time = query_time


class ANALYTIC_CONDITION(Protocol):
	def __call__(
		self,
		skip_events: 'NAWAH_EVENTS',
		env: 'NAWAH_ENV',
		query: Union['Query', 'NAWAH_QUERY'],
		doc: 'NAWAH_DOC',
		method: str,
	) -> bool:
		...


class ANALYTIC_DOC(Protocol):
	def __call__(
		self,
		skip_events: 'NAWAH_EVENTS',
		env: 'NAWAH_ENV',
		query: Union['Query', 'NAWAH_QUERY'],
		doc: 'NAWAH_DOC',
		method: str,
	) -> 'NAWAH_DOC':
		...


class ANALYTIC:
	condition: ANALYTIC_CONDITION
	doc: ANALYTIC_DOC

	def __init__(
		self,
		*,
		condition: ANALYTIC_CONDITION,
		doc: ANALYTIC_DOC,
	):
		setattr(self, 'condition', condition)
		setattr(self, 'doc', doc)



import nawah
from nawah.config import Config, process_config
from nawah.test import TEST
from nawah.classes import L10N, ATTR
from nawah.utils import validate_attr

import logging, sys, os.path, pkgutil, inspect, re

logger = logging.getLogger('nawah')


async def import_modules():
	sys.path.append(os.path.join(nawah.__path__[0], 'packages'))
	sys.path.append(os.path.join(Config._app_path, 'packages'))

	# [DOC] Iterate over packages in modules folder
	for _, pkgname, ispkg in pkgutil.iter_modules(
		[
			os.path.join(nawah.__path__[0], 'packages'),
			os.path.join(Config._app_path, 'packages'),
		]
	):
		if not ispkg:
			continue

		logger.debug(f'Importing package: {pkgname}')

		# [DOC] Load package and attempt to load config
		package = __import__(pkgname, fromlist='*')
		process_config(config=package.config, pkgname=pkgname)

		# [DOC] Add package to loaded packages dict
		Config.modules_packages[pkgname] = []

		# [DOC] Iterate over python modules in package
		package_prefix = package.__name__ + '.'
		for _, modname, ispkg in pkgutil.iter_modules(package.__path__, package_prefix):
			# [DOC] Iterate over python classes in module
			module = __import__(modname, fromlist='*')
			if modname.endswith('__tests__'):
				for test_name in dir(module):
					if type(getattr(module, test_name)) == TEST:
						Config.tests[test_name] = getattr(module, test_name)
				continue
			elif modname.endswith('__l10n__'):
				for l10n_name in dir(module):
					if type(getattr(module, l10n_name)) == L10N:
						if l10n_name not in Config.l10n.keys():
							Config.l10n[l10n_name] = {}
						Config.l10n[l10n_name].update(getattr(module, l10n_name))
				continue

			for clsname in dir(module):
				# [DOC] Confirm class is subclass of BaseModule
				if (
					clsname != 'BaseModule'
					and inspect.isclass(getattr(module, clsname))
					and getattr(getattr(module, clsname), '_nawah_module', False)
				):
					# [DOC] Deny loading Nawah-reserved named Nawah modules
					if clsname.lower() in ['conn', 'heart', 'watch']:
						logger.error(
							f'Module with Nawah-reserved name \'{clsname.lower()}\' was found. Exiting.'
						)
						exit(1)
					# [DOC] Load Nawah module and assign module_name attr
					cls = getattr(module, clsname)
					module_name = re.sub(r'([A-Z])', r'_\1', clsname[0].lower() + clsname[1:]).lower()
					# [DOC] Deny duplicate Nawah modules names
					if module_name in Config.modules.keys():
						logger.error(f'Duplicate module name \'{module_name}\'. Exiting.')
						exit(1)
					# [DOC] Add module to loaded modules dict
					Config.modules[module_name] = cls()
					Config.modules_packages[pkgname].append(module_name)

	# [DOC] Update User, Session modules with populated attrs
	Config.modules['user'].attrs.update(Config.user_attrs)
	if (
		sum(1 for attr in Config.user_settings.keys() if attr in Config.user_attrs.keys())
		!= 0
	):
		logger.error(
			'At least on attr from \'user_settings\' is conflicting with an attr from \'user_attrs\'. Exiting.'
		)
		exit(1)
	Config.modules['user'].defaults['locale'] = Config.locale
	for attr in Config.user_attrs.keys():
		Config.modules['user'].unique_attrs.append(attr)
		Config.modules['user'].attrs[f'{attr}_hash'] = ATTR.STR()
		Config.modules['session'].methods['auth'].doc_args.append(
			{
				'hash': ATTR.STR(),
				attr: Config.user_attrs[attr],
				'groups': ATTR.LIST(list=[ATTR.ID()]),
			}
		)
		Config.modules['session'].methods['auth'].doc_args.append(
			{'hash': ATTR.STR(), attr: Config.user_attrs[attr]}
		)

	# [DOC] Attempt to validate all packages required vars (via vars_types Config Attr) are met
	for var in Config.vars_types.keys():
		if var not in Config.vars.keys():
			logger.error(
				f'Package \'{Config.vars_types[var]["package"]}\' requires \'{var}\' Var, but not found in App Config. Exiting.'
			)
			exit(1)
		try:
			await validate_attr(
				mode='create',
				attr_name=var,
				attr_type=Config.vars_types[var]['type'],
				attr_val=Config.vars[var],
			)
		except:
			logger.error(
				f'Package \'{Config.vars_types[var]["package"]}\' requires \'{var}\' Var of type \'{Config.vars_types[var]["type"]._type}\', but validation failed. Exiting.'
			)
			exit(1)

	# [DOC] Call update_modules, effectively finalise initialising modules
	for module in Config.modules.values():
		module._initialise()
	# [DOC] Write api_ref if generate_ref mode
	if Config.generate_ref:
		from nawah.utils import generate_ref

		generate_ref()
	# [DOC] Write api_models if generate_models mode
	if Config.generate_models:
		from nawah.utils import generate_models

		generate_models()


from nawah.utils import validate_attr, InvalidAttrException, ConvertAttrException
from nawah.classes import (
	MethodException,
	DictObj,
	BaseModel,
	Query,
	JSONEncoder,
	ATTR,
	ATTR_MOD,
	NAWAH_EVENTS,
	NAWAH_ENV,
	NAWAH_DOC,
	NAWAH_QUERY,
	PERM,
	WATCH_TASK,
	METHOD,
)
from nawah.enums import Event, NAWAH_VALUES
from nawah.config import Config

from asyncio import coroutine
from aiohttp.web import WebSocketResponse
from typing import (
	List,
	Dict,
	Union,
	Any,
	Tuple,
	Set,
	AsyncGenerator,
	TYPE_CHECKING,
	Literal,
	Iterable,
	Optional,
	cast,
	Awaitable,
)

import logging, copy, traceback, sys, asyncio

if TYPE_CHECKING:
	from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


class BaseMethod:
	def __init__(
		self,
		module: 'BaseModule',
		method: str,
		permissions: List[PERM],
		query_args: List[Dict[str, ATTR]],
		doc_args: List[Dict[str, ATTR]],
		watch_method: bool,
		get_method: bool,
		post_method: bool,
	):
		self.module = module
		self.method = method
		self.permissions = permissions
		self.query_args = query_args
		self.doc_args = doc_args
		self.watch_method = watch_method
		self.get_method = get_method
		self.post_method = post_method

	async def validate_args(
		self, args: Union[Query, Dict[str, Any]], args_list_label: str
	):
		args_list = getattr(self, f'{args_list_label}_args')

		sets_check: List[Dict[str, Literal[True, 'missing', 'invalid', 'convert']]] = []

		for args_set in args_list:
			set_status = True
			set_check = len(sets_check)
			sets_check.append({arg: True for arg in args_set.keys()})

			if args_list_label == 'query':
				args_check: Iterable = args
			elif args_list_label == 'doc':
				args = cast(Dict[str, Any], args)
				args_check = args.keys()

			for arg in args_set.keys():
				if arg not in args_check:
					set_status = False
					sets_check[set_check][arg] = 'missing'
				else:
					try:
						if args_list_label == 'query' and arg[0] != '$':
							for i in range(len(args[arg])):
								if (
									':' in arg
									and type(args[arg][i]) == dict
									and (arg_oper := arg.split(':')[1]) in args[arg][i].keys()
								):
									args[arg][i][arg_oper] = await validate_attr(
										mode='create',
										attr_name=arg,
										attr_type=args_set[arg],
										attr_val=args[arg][i][arg_oper],
									)
								else:
									args[arg][i] = await validate_attr(
										mode='create',
										attr_name=arg,
										attr_type=args_set[arg],
										attr_val=args[arg][i],
									)
						elif args_list_label == 'query' and arg[0] == '$':
							args[arg] = await validate_attr(
								mode='create',
								attr_name=arg,
								attr_type=args_set[arg],
								attr_val=args[arg],
							)
						elif args_list_label == 'doc':
							args[arg] = await validate_attr(
								mode='create',
								attr_name=arg,
								attr_type=args_set[arg],
								attr_val=args[arg],
							)
					except InvalidAttrException:
						set_status = False
						sets_check[set_check][arg] = 'invalid'
					except ConvertAttrException:
						set_status = False
						sets_check[set_check][arg] = 'convert'

			if set_status:
				return True

		return sets_check

	async def __call__(
		self,
		*,
		skip_events: NAWAH_EVENTS = None,
		env: NAWAH_ENV = None,
		query: Union[NAWAH_QUERY, Query] = None,
		doc: NAWAH_DOC = None,
		call_id: str = None,
	) -> Optional[DictObj]:
		if skip_events == None:
			skip_events = []
		if env == None:
			env = {}
		if query == None:
			query = []
		if doc == None:
			doc = {}
		skip_events = cast(NAWAH_EVENTS, skip_events)
		env = cast(NAWAH_ENV, env)
		query = cast(Union[NAWAH_QUERY, Query], query)
		doc = cast(NAWAH_DOC, doc)
		call_id = cast(str, call_id)
		# [DOC] Convert list query to Query object
		query = Query(copy.deepcopy(query))
		# [DOC] deepcopy() doc object ro prevent mutating original doc
		doc = copy.deepcopy(doc)

		logger.debug(
			f'Calling: {self.module.module_name}.{self.method}, with skip_events:{skip_events}, query:{str(query)[:250]}, doc.keys:{doc.keys()}'
		)

		if call_id:
			for analytics_set in self.module.analytics:
				if analytics_set.condition(
					skip_events=skip_events,
					env=env,
					query=query,
					doc=doc,
					method=self.method,
				):
					try:
						analytic_doc = analytics_set.doc(
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							method=self.method,
						)
						analytic_results = await Config.modules['analytic'].create(
							skip_events=[Event.PERM], env=env, doc=analytic_doc
						)
					except Exception as e:
						logger.error(
							f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
						)
					if analytic_results.status != 200:
						logger.error(
							f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
						)

		if Event.ARGS not in skip_events and Config.realm:
			if self.module.module_name == 'realm':
				if self.method != 'create':
					query.append({'name': env['realm']})
					doc['name'] = env['realm']
					logger.debug(
						f'Appended realm name attrs to query, doc: {str(query)[:250]}, {doc.keys()}'
					)
				else:
					logger.debug(
						'Skipped Appending realm name attrs to query, doc for realm.create call'
					)
			else:
				query.append({'realm': env['realm']})
				doc['realm'] = env['realm']
				logger.debug(
					f'Appended realm attrs to query, doc: {JSONEncoder().encode(query)}, {doc.keys()}'
				)

		if Event.PERM not in skip_events and env['session']:
			permissions_check = Config.modules['session'].check_permissions(
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				module=self.module,
				permissions=self.permissions,
			)
			logger.debug(f'permissions_check: {permissions_check}.')
			if permissions_check == False:
				return await self.return_results(
					ws=env['ws'] if 'ws' in env.keys() else None,
					results=DictObj(
						{
							'status': 403,
							'msg': 'You don\'t have permissions to access this endpoint.',
							'args': DictObj({'code': 'CORE_SESSION_FORBIDDEN'}),
						}
					),
					call_id=call_id,
				)
			else:
				if type(permissions_check['query']) == dict:
					permissions_check['query'] = [permissions_check['query']]
				for i in range(len(permissions_check['query'])):
					del_args = []
					# [DOC] attempt to process query_set as nested-list (OR) even if it's dict
					if type(permissions_check['query'][i]) == dict:
						query_set_list = [permissions_check['query'][i]]
					elif type(permissions_check['query'][i]) == list:
						query_set_list = permissions_check['query'][i]
					# [DOC] loop over query_set_list, query_set
					for query_set in query_set_list:
						for attr in query_set.keys():
							# [DOC] Flag attr for deletion if value is None
							if query_set[attr] == None or type(query_set[attr]) == ATTR_MOD:
								del_args.append(attr)
						for attr in del_args:
							del query_set[attr]
				# [DOC] Append query permissions args to query
				query.append(permissions_check['query'])

				del_args = []
				for attr in permissions_check['doc'].keys():
					# [DOC] Replace None value with NONE_VALUE to bypass later validate step
					if permissions_check['doc'][attr] == None:
						permissions_check['doc'][attr] = NAWAH_VALUES.NONE_VALUE
				for attr in del_args:
					del permissions_check['doc'][attr]
				# [DOC] Update doc with doc permissions args
				doc.update(permissions_check['doc'])
				doc = {
					attr: doc[attr] for attr in doc.keys() if doc[attr] != NAWAH_VALUES.NONE_VALUE
				}

		if Event.ARGS not in skip_events:
			if self.query_args:
				test_query = await self.validate_args(query, 'query')
				if test_query != True:
					for i in range(len(test_query)):
						test_query[i] = (
							'['
							+ ', '.join(
								[
									f'\'{arg}\': {val.capitalize()}'
									for arg, val in test_query[i].items()
									if val != True
								]
							)
							+ ']'
						)
					return await self.return_results(
						ws=env['ws'] if 'ws' in env.keys() else None,
						results=DictObj(
							{
								'status': 400,
								'msg': 'Could not match query with any of the required query_args. Failed sets:'
								+ ', '.join(test_query),
								'args': DictObj(
									{
										'code': f'{self.module.package_name.upper()}_{self.module.module_name.upper()}_INVALID_QUERY'
									}
								),
							}
						),
						call_id=call_id,
					)

			if self.doc_args:
				test_doc = await self.validate_args(doc, 'doc')
				if test_doc != True:
					for i in range(len(test_doc)):
						test_doc[i] = (
							'['
							+ ', '.join(
								[
									f'\'{arg}\': {val.capitalize()}'
									for arg, val in test_doc[i].items()
									if val != True
								]
							)
							+ ']'
						)
					return await self.return_results(
						ws=env['ws'] if 'ws' in env.keys() else None,
						results=DictObj(
							{
								'status': 400,
								'msg': 'Could not match doc with any of the required doc_args. Failed sets:'
								+ ', '.join(test_doc),
								'args': DictObj(
									{
										'code': f'{self.module.package_name.upper()}_{self.module.module_name.upper()}_INVALID_DOC'
									}
								),
							}
						),
						call_id=call_id,
					)

		for arg in doc.keys():
			if type(doc[arg]) == BaseModel:
				doc[arg] = doc[arg]._id  # type: ignore

		# [DOC] check if $soft oper is set to add it to events
		if '$soft' in query and query['$soft'] == True:
			skip_events.append(Event.SOFT)
			del query['$soft']

		# [DOC] check if $extn oper is set to add it to events
		if '$extn' in query and query['$extn'] == False:
			skip_events.append(Event.EXTN)
			del query['$extn']

		try:
			# [DOC] Check for proxy module
			if self.module.proxy:
				if not getattr(self.module, f'_method_{self.method}', None):
					method = getattr(Config.modules[self.module.proxy], f'_method_{self.method}')
				else:
					method = getattr(self.module, f'_method_{self.method}')
			else:
				method = getattr(self.module, f'_method_{self.method}')
			# [DOC] Call method function
			if self.watch_method:
				await env['ws'].send_str(
					JSONEncoder().encode(
						{
							'status': 200,
							'msg': 'Created watch task.',
							'args': {
								'code': 'CORE_WATCH_OK',
								'watch': call_id,
								'call_id': call_id,
							},
						}
					)
				)
				watch_loop = self.watch_loop(
					ws=env['ws'],
					stream=method(skip_events=skip_events, env=env, query=query, doc=doc),
					call_id=call_id,
					watch_task=env['watch_tasks'][call_id],
				)
				env['watch_tasks'][call_id] = {'watch': watch_loop}
				env['watch_tasks'][call_id]['task'] = asyncio.create_task(watch_loop)
				return None
			else:
				try:
					results = await method(skip_events=skip_events, env=env, query=query, doc=doc)
				except MethodException as e:
					results = e.args[0]

				if type(results) == coroutine:
					raise TypeError('Method returned coroutine rather than acceptable results format.')

				results = DictObj(results)
				try:
					results['args'] = DictObj(results.args)
				except Exception:
					results['args'] = DictObj({})

				logger.debug(f'Call results: {JSONEncoder().encode(results)}')
				# [DOC] Check for session in results
				if 'session' in results.args:
					if results.args.session._id == 'f00000000000000000000012':
						# [DOC] Updating session to __ANON
						anon_user = Config.compile_anon_user()
						anon_session = Config.compile_anon_session()
						anon_session['user'] = DictObj(anon_user)
						env['session'] = BaseModel(anon_session)
					else:
						# [DOC] Updating session to user
						env['session'] = results.args.session

				return await self.return_results(
					ws=env['ws'] if 'ws' in env.keys() else None, results=results, call_id=call_id
				)
			# query = Query([])
		except Exception as e:
			logger.error(f'An error occurred. Details: {traceback.format_exc()}.')
			tb = sys.exc_info()[2]
			if tb is not None:
				prev = tb
				current = tb.tb_next
				while current is not None:
					prev = current
					current = current.tb_next
				logger.error(f'Scope variables: {JSONEncoder().encode(prev.tb_frame.f_locals)}')
			query = Query([])
			if Config.debug:
				return await self.return_results(
					ws=env['ws'] if 'ws' in env.keys() else None,
					results=DictObj(
						{
							'status': 500,
							'msg': f'Unexpected error has occurred [method:{self.module.module_name}.{self.method}] [{str(e)}].',
							'args': DictObj(
								{
									'code': 'CORE_SERVER_ERROR',
									'method': f'{self.module.module_name}.{self.method}',
									'err': str(e),
								}
							),
						}
					),
					call_id=call_id,
				)
			else:
				return await self.return_results(
					ws=env['ws'] if 'ws' in env.keys() else None,
					results=DictObj(
						{
							'status': 500,
							'msg': 'Unexpected error has occurred.',
							'args': DictObj({'code': 'CORE_SERVER_ERROR'}),
						}
					),
					call_id=call_id,
				)

	async def return_results(
		self, ws: Optional[WebSocketResponse], results: DictObj, call_id: Optional[str]
	) -> Optional[DictObj]:
		if call_id and call_id != '__TEST__':
			results.args['call_id'] = call_id
			ws = cast(WebSocketResponse, ws)
			await ws.send_str(JSONEncoder().encode(results))
			return None
		else:
			return results

	async def watch_loop(
		self,
		ws: WebSocketResponse,
		stream: AsyncGenerator[DictObj, DictObj],
		call_id: str,
		watch_task: WATCH_TASK,
	) -> None:
		logger.debug('Preparing async loop at BaseMethod')
		async for results in stream:
			logger.debug(f'Received watch results at BaseMethod: {results}')
			# [DOC] Update watch_task stream value with stream object
			if 'stream' in results.keys():
				watch_task['stream'] = results['stream']
				continue

			results = DictObj(results)
			try:
				results['args'] = DictObj(results.args)
			except Exception:
				results['args'] = DictObj({})

			results.args['call_id'] = call_id
			results.args['watch'] = call_id

			await ws.send_str(JSONEncoder().encode(results))

		logger.debug('Generator ended at BaseMethod.')



from nawah.enums import Event, LOCALE_STRATEGY
from nawah.classes import (
	DictObj,
	BaseModel,
	NAWAH_DOC,
	NAWAH_ENV,
	ATTR,
	APP_CONFIG,
	PACKAGE_CONFIG,
	CLIENT_APP,
	ANALYTICS_EVENTS,
	SYS_DOC,
	JOB,
)

from typing import (
	List,
	Dict,
	Callable,
	Any,
	Union,
	Set,
	Tuple,
	Literal,
	TypedDict,
	Optional,
	cast,
	TYPE_CHECKING,
)

from croniter import croniter
from motor.motor_asyncio import AsyncIOMotorClient
from bson import ObjectId
from passlib.hash import pbkdf2_sha512

import os, logging, datetime, time, requests

if TYPE_CHECKING:
	from nawah.test import STEP
	from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


def process_config(*, config: Union[APP_CONFIG, PACKAGE_CONFIG], pkgname: str = None):
	from nawah.utils import deep_update

	if type(config) not in [APP_CONFIG, PACKAGE_CONFIG]:
		logger.error(f'Config object of type \'{type(config)}\' is invalid. Exiting.')
		exit(1)

	if type(config) == PACKAGE_CONFIG and not pkgname:
		logger.error(
			'Provided Config object of type \'PACKAGE_CONFIG\' without \'pkgname\'. Exiting.'
		)
		exit(1)

	app_only_attrs = APP_CONFIG.__annotations__.keys()

	for config_attr in dir(config):
		config_attr_val = getattr(config, config_attr)

		# [DOC] Check existence of of api_level, version Config Attrs
		if type(config) == PACKAGE_CONFIG and config_attr in ['api_level', 'version']:
			if config_attr_val == None:
				logger.error(
					f'Package \'{pkgname}\' is missing \'{config_attr}\' Config Attr. Exiting.'
				)
				exit(1)
			# [DOC] Check type of api_level, version Config Attrs
			elif type(config_attr_val) != str:
				logger.error(
					f'Package \'{pkgname}\' is having invalid type of \'{config_attr}\'. Exiting.'
				)
				exit(1)
			else:
				# [DOC] Update corresponding Config
				getattr(Config, f'packages_{config_attr}s')[pkgname] = config_attr_val
		# [DOC] Skip non Config Attr attrs
		elif (
			config_attr.startswith('__')
			or config_attr_val == None
			or config_attr in app_only_attrs
		):
			continue
		# [DOC] For vars_types Config Attr, preserve package name for debugging purposes
		elif config_attr == 'vars_types':
			for var in config_attr_val.keys():
				Config.vars_types[var] = {'package': pkgname, 'type': config_attr_val[var]}
		elif type(config_attr_val) == list:
			for j in config_attr_val:
				getattr(Config, config_attr).append(j)
			if config_attr == 'locales':
				Config.locales = list(set(Config.locales))
		elif type(config_attr_val) == dict:
			if not getattr(Config, config_attr):
				setattr(Config, config_attr, {})
			deep_update(target=getattr(Config, config_attr), new_values=config_attr_val)
		else:
			setattr(Config, config_attr, config_attr_val)


class Config:
	debug: bool = False
	env: str
	port: int = 8081

	_sys_conn: AsyncIOMotorClient
	_sys_env: NAWAH_ENV
	_sys_docs: Dict[ObjectId, SYS_DOC] = {}
	_realms: Dict[str, BaseModel] = {}
	_jobs_base: datetime.datetime

	_nawah_version: str
	packages_api_levels: Dict[str, str] = {}
	packages_versions: Dict[str, str] = {}

	_app_name: str
	_app_version: str
	_app_default_package: str
	_app_path: str
	_app_packages: Dict[str, str]

	test: bool = False
	test_name: str
	test_skip_flush: bool = False
	test_force: bool = False
	test_env: bool = False
	test_breakpoint: bool = False
	test_collections: bool = False
	tests: Dict[str, List['STEP']] = {}

	emulate_test: bool = False
	force_admin_check: bool = False

	generate_ref: bool = False
	_api_ref: str

	generate_models: bool = False
	_api_models: str

	realm: bool = False

	vars_types: Dict[str, Union[ATTR, Dict[str, Any]]] = {}
	vars: Dict[str, Any] = {}

	client_apps: Dict[str, CLIENT_APP] = {}

	analytics_events: ANALYTICS_EVENTS = {
		'app_conn_verified': True,
		'session_conn_auth': True,
		'session_user_auth': True,
		'session_conn_reauth': True,
		'session_user_reauth': True,
		'session_conn_deauth': True,
		'session_user_deauth': True,
	}

	conn_timeout: int = 120
	quota_anon_min: int = 40
	quota_auth_min: int = 100
	quota_ip_min: int = 500
	file_upload_limit: int = -1
	file_upload_timeout: int = 300

	data_server: str = 'mongodb://localhost'
	data_name: str = 'nawah_data'
	data_ssl: bool = False
	data_ca_name: Optional[str] = None
	data_ca: Optional[str] = None
	data_disk_use: bool = False

	data_azure_mongo: bool = False

	locales: List[str] = ['ar_AE', 'en_AE']
	locale: str = 'ar_AE'
	locale_strategy: LOCALE_STRATEGY = LOCALE_STRATEGY.DUPLICATE

	admin_doc: NAWAH_DOC = {}
	admin_password: str = '__0xADMIN'

	anon_token: str = '__ANON_TOKEN_f00000000000000000000012'
	anon_privileges: Dict[str, List[str]] = {}

	user_attrs: Dict[str, ATTR] = {}
	user_settings: Dict[
		str, Dict[Literal['type', 'val'], Union[Literal['user', 'user_sys'], Any]]
	] = {}
	user_doc_settings: List[str] = []

	groups: List[Dict[str, Any]] = []
	default_privileges: Dict[str, List[str]] = {}

	data_indexes: List[Dict[str, Any]] = []

	docs: List[SYS_DOC] = []

	l10n: Dict[str, Dict[str, Any]] = {}

	jobs: Dict[str, JOB] = {}

	gateways: Dict[str, Callable] = {}

	types: Dict[str, Callable] = {}

	modules: Dict[str, 'BaseModule'] = {}
	modules_packages: Dict[str, List[str]] = {}

	@classmethod
	async def config_data(cls) -> None:
		from nawah.utils import generate_attr

		# [TODO] Add validator for user_attrs, user_settings, user_doc_settings

		# [DOC] Check app packages
		if cls._app_packages or len(cls.packages_versions.keys()) > 2:
			logger.debug(
				'Found \'_app_packages\' Config Attr. Attempting to validate all loaded packages are matching _app_packages Config Attr value.'
			)

			cls._app_packages['core'] = cls.packages_versions['core']
			cls._app_packages[cls._app_default_package] = cls.packages_versions[
				cls._app_default_package
			]

			missing_packages = [
				package
				for package in cls._app_packages.keys()
				if package not in cls.packages_versions.keys()
			]
			if missing_packages:
				logger.error(
					f'At least one package is missing that is required by app. Missing package[s]: \'{", ".join(missing_packages)}\'. Exiting.'
				)
				exit(1)

			extra_packages = [
				package
				for package in cls.packages_versions.keys()
				if package not in cls._app_packages.keys()
			]
			if extra_packages:
				logger.error(
					f'At least one extra package is present in \'packages\' folder that is not required by app. Extra package[s]: \'{", ".join(extra_packages)}\'. Exiting.'
				)
				exit(1)

			# [DOC] Check for version mismatch
			for package, version in cls._app_packages.items():
				# [DOC] Skip core and default_packages
				if package in ['core', cls._app_default_package]:
					continue
				if version != cls.packages_versions[package]:
					logger.error(
						f'Package \'{package}\' version \'{cls.packages_versions[package]}\' is added to app but not matching required version \'{version}\'. Exiting.'
					)
					exit(1)

		# [DOC] Check API version
		if not cls.packages_api_levels:
			logger.warning(
				'No API-level specified for the app. Nawah would continue to run the app, but the developer should consider adding API-level to eliminate specs mismatch.'
			)
		else:
			nawah_level = '.'.join(cls._nawah_version.split('.')[0:2])
			for package, api_level in cls.packages_api_levels.items():
				if api_level != nawah_level:
					logger.error(
						f'Nawah framework is on API-level \'{nawah_level}\', but the app package \'{package}\' requires API-level \'{api_level}\'. Exiting.'
					)
					exit(1)
			try:
				versions = (
					(
						requests.get(
							'https://raw.githubusercontent.com/masaar/nawah_versions/master/versions.txt'
						).content
					)
					.decode('utf-8')
					.split('\n')
				)
				version_detected = ''
				for version in versions:
					if version.startswith(f'{nawah_level}.'):
						if version_detected and int(version.split('.')[-1]) < int(
							version_detected.split('.')[-1]
						):
							continue
						version_detected = version
				if version_detected and version_detected != cls._nawah_version:
					logger.warning(
						f'Your app is using Nawah version \'{cls._nawah_version}\' while newer version \'{version_detected}\' of the API-level is available. Please, update.'
					)
			except:
				logger.warning(
					'An error occurred while attempting to check for latest update to Nawah. Please, check for updates on your own.'
				)

		# [DOC] Check for jobs
		if cls.jobs:
			# [DOC] Check jobs schedule validity
			cls._jobs_base = datetime.datetime.utcnow()
			for job_name in cls.jobs.keys():
				job = cls.jobs[job_name]
				if not croniter.is_valid(job.schedule):
					logger.error(f'Job with schedule \'{job_name}\' schedule is invalid. Exiting.')
					exit(1)

				job._cron_schedule = croniter(job.schedule, cls._jobs_base)
				job._next_time = datetime.datetime.fromtimestamp(
					job._cron_schedule.get_next(), datetime.timezone.utc
				).isoformat()[:16]

		# [DOC] Check for presence of user_auth_attrs
		if not cls.user_attrs.keys():
			logger.error('No \'user_attrs\' are provided. Exiting.')
			exit(1)

		# [DOC] Check default values
		security_warning = '[SECURITY WARNING] {config_attr} is not explicitly set. It has been defaulted to \'{val}\' but in production environment you should consider setting it to your own to protect your app from breaches.'
		if cls.admin_password == '__ADMIN':
			logger.warning(security_warning.format(config_attr='Admin password', val='__ADMIN'))
		if cls.anon_token == '__ANON_TOKEN_f00000000000000000000012':
			logger.warning(
				security_warning.format(
					config_attr='Anon token',
					val='__ANON_TOKEN_f00000000000000000000012',
				)
			)

		# [DOC] Check for Env Vars
		attrs_defaults = {
			'data_server': 'mongodb://localhost',
			'data_name': 'nawah_data',
			'data_ssl': False,
			'data_ca_name': False,
			'data_ca': False,
			'emulate_test': False,
		}
		for attr_name in attrs_defaults.keys():
			attr_val = getattr(cls, attr_name)
			if type(attr_val) == str and attr_val.startswith('$__env.'):
				logger.debug(f'Detected Env Variable for config attr \'{attr_name}\'')
				if not os.getenv(attr_val[7:]):
					logger.warning(
						f'Couldn\'t read Env Variable for config attr \'{attr_name}\'. Defaulting to \'{attrs_defaults[attr_name]}\''
					)
					setattr(cls, attr_name, attrs_defaults[attr_name])
				else:
					# [DOC] Set data_ssl to True rather than string Env Variable value
					if attr_name == 'ssl':
						attr_val = True
					else:
						attr_val = os.getenv(attr_val[7:])
					logger.warning(
						f'Setting Env Variable for config attr \'{attr_name}\' to \'{attr_val}\''
					)
					setattr(cls, attr_name, attr_val)

		# [DOC] Check SSL settings
		if cls.data_ca and cls.data_ca_name:
			__location__ = os.path.realpath(os.path.join('.'))
			if not os.path.exists(os.path.join(__location__, 'certs')):
				os.makedirs(os.path.join(__location__, 'certs'))
			with open(os.path.join(__location__, 'certs', cls.data_ca_name), 'w') as f:
				f.write(cls.data_ca)

		from nawah import data as Data

		# [DOC] Create default env dict
		anon_user = cls.compile_anon_user()
		anon_session = DictObj(cls.compile_anon_session())
		anon_session = cast(BaseModel, anon_session)
		anon_session['user'] = DictObj(anon_user)
		cls._sys_conn = Data.create_conn()
		cls._sys_env = {
			'conn': cls._sys_conn,
			'REMOTE_ADDR': '127.0.0.1',
			'HTTP_USER_AGENT': 'Nawah',
			'client_app': '__sys',
			'session': anon_session,
			'watch_tasks': {},
		}

		if cls.data_azure_mongo:
			for module in cls.modules.keys():
				try:
					if cls.modules[module].collection:
						logger.debug(
							f'Attempting to create shard collection: {cls.modules[module].collection}.'
						)
						cls._sys_conn[cls.data_name].command(
							'shardCollection',
							f'{cls.data_name}.{cls.modules[module].collection}',
							key={'_id': 'hashed'},
						)
					else:
						logger.debug(f'Skipping service module: {module}.')
				except Exception as err:
					logger.error(err)

		# [DOC] Check test mode
		if cls.test or cls.test_collections:
			logger.debug('Test mode or Test Collections Mode detected.')
			__location__ = os.path.realpath(os.path.join('.'))
			if not os.path.exists(os.path.join(__location__, 'tests')):
				os.makedirs(os.path.join(__location__, 'tests'))
			if not cls.test_env:
				for module in cls.modules.keys():
					module_collection = cls.modules[module].collection
					if module_collection:
						logger.debug(
							f'Updating collection name \'{module_collection}\' of module {module}'
						)
						module_collection = cls.modules[module].collection = f'test_{module_collection}'
						if cls.test and not cls.test_skip_flush:
							logger.debug(f'Flushing test collection \'{module_collection}\'')
							await Data.drop(
								env=cls._sys_env,
								collection_name=module_collection,
							)
					else:
						logger.debug(f'Skipping service module {module}')
			else:
				logger.warning(
					f'Testing on \'{cls.env}\' env. Nawah would be sleeping for 5sec to give you chance to abort test workflow if this was a mistake.'
				)
				time.sleep(5)

		logger.debug('Testing realm mode.')
		if cls.realm:
			# [DOC] Append realm to env dict
			cls._sys_env['realm'] = '__global'
			# [DOC] Append realm attrs to all modules attrs and set at as required in query_args and doc_args
			for module in cls.modules.keys():
				if module != 'realm':
					logger.debug(f'Updating module \'{module}\' for realm mode.')
					cls.modules[module].attrs['realm'] = ATTR.STR()
					for method in cls.modules[module].methods.keys():
						module_method = cls.modules[module].methods[method]
						# [DOC] Attempt required changes to query_args to add realm query_arg
						for query_args_set in cast(List[Dict[str, ATTR]], module_method.query_args):
							query_args_set['realm'] = ATTR.STR()
						# [DOC] Attempt required changes to doc_args to add realm doc_arg
						for doc_args_set in cast(List[Dict[str, ATTR]], module_method.doc_args):
							doc_args_set['realm'] = ATTR.STR()
						module_method._callable.query_args = cast(
							List[Dict[str, ATTR]], module_method.query_args
						)
						module_method._callable.doc_args = cast(
							List[Dict[str, ATTR]], module_method.doc_args
						)
			# [DOC] Query all realms to provide access to available realms and to add realm docs to _sys_docs
			realm_results = await cls.modules['realm'].read(
				skip_events=[Event.PERM, Event.ARGS], env=cls._sys_env
			)
			logger.debug(
				f'Found {realm_results.args.count} realms. Namely; {[doc.name for doc in realm_results.args.docs]}'
			)
			for doc in realm_results.args.docs:
				cls._realms[doc.name] = doc
				cls._sys_docs[doc._id] = SYS_DOC(module='realm')
			# [DOC] Create __global realm
			if '__global' not in cls._realms:
				logger.debug('GLOBAL realm not found, creating it.')
				realm_results = await cls.modules['realm'].create(
					skip_events=[Event.PERM, Event.PRE],
					env=cls._sys_env,
					doc={
						'_id': ObjectId('f00000000000000000000014'),
						'user': ObjectId('f00000000000000000000010'),
						'name': '__global',
						'default': 'f00000000000000000000013',
					},
				)
				logger.debug(f'GLOBAL realm creation results: {realm_results}')
				if realm_results.status != 200:
					logger.error('Config step failed. Exiting.')
					exit(1)

		# [DOC] Checking users collection
		# [TODO] Updated sequence to handle users
		logger.debug('Testing users collection.')
		user_results = await cls.modules['user'].read(
			skip_events=[Event.PERM, Event.ON],
			env=cls._sys_env,
			query=[{'_id': 'f00000000000000000000010'}],
		)
		if not user_results.args.count:
			logger.debug('ADMIN user not found, creating it.')
			# [DOC] Prepare base ADMIN user doc
			admin_create_doc = {
				'_id': ObjectId('f00000000000000000000010'),
				'name': {cls.locale: '__ADMIN'},
				'groups': [],
				'privileges': {'*': ['*']},
				'locale': cls.locale,
			}
			# [DOC] Update ADMIN user doc with admin_doc Config Attr
			admin_create_doc.update(cls.admin_doc)

			for auth_attr in cls.user_attrs.keys():
				admin_create_doc[f'{auth_attr}_hash'] = pbkdf2_sha512.using(rounds=100000).hash(
					f'{auth_attr}{admin_create_doc[auth_attr]}{cls.admin_password}{cls.anon_token}'.encode(
						'utf-8'
					)
				)
			if cls.realm:
				admin_create_doc['realm'] = '__global'
			admin_results = await cls.modules['user'].create(
				skip_events=[Event.PERM],
				env=cls._sys_env,
				doc=admin_create_doc,
			)
			logger.debug(f'ADMIN user creation results: {admin_results}')
			if admin_results.status != 200:
				logger.error('Config step failed. Exiting.')
				exit(1)
		elif not cls.force_admin_check:
			logger.warning(
				'ADMIN user found, skipping check due to force_admin_check Config Attr.'
			)
		else:
			logger.warning('ADMIN user found, checking it due to force_admin_check Config Attr.')
			admin_doc: BaseModel = user_results.args.docs[0]
			admin_doc_update = {}
			for attr in cls.admin_doc.keys():
				if (
					attr not in admin_doc
					or not admin_doc[attr]
					or cls.admin_doc[attr] != admin_doc[attr]
				):
					if (
						type(cls.admin_doc[attr]) == dict
						and cls.locale in cls.admin_doc[attr].keys()
						and type(admin_doc[attr]) == dict
						and (
							(
								cls.locale in admin_doc[attr].keys()
								and cls.admin_doc[attr][cls.locale] == admin_doc[attr][cls.locale]
							)
							or (cls.locale not in admin_doc[attr].keys())
						)
					):
						continue
					logger.debug(f'Detected change in \'admin_doc.{attr}\' Config Attr.')
					admin_doc_update[attr] = cls.admin_doc[attr]
			for auth_attr in cls.user_attrs.keys():
				auth_attr_hash = pbkdf2_sha512.using(rounds=100000).hash(
					f'{auth_attr}{admin_doc[auth_attr]}{cls.admin_password}{cls.anon_token}'.encode(
						'utf-8'
					)
				)
				if (
					f'{auth_attr}_hash' not in admin_doc
					or auth_attr_hash != admin_doc[f'{auth_attr}_hash']
				):
					logger.debug(f'Detected change in \'admin_password\' Config Attr.')
					admin_doc_update[f'{auth_attr}_hash'] = auth_attr_hash
			if len(admin_doc_update.keys()):
				logger.debug(f'Attempting to update ADMIN user with doc: \'{admin_doc_update}\'')
				admin_results = await cls.modules['user'].update(
					skip_events=[Event.PERM, Event.PRE, Event.ON],
					env=cls._sys_env,
					query=[{'_id': ObjectId('f00000000000000000000010')}],
					doc=admin_doc_update,
				)
				logger.debug(f'ADMIN user update results: {admin_results}')
				if admin_results.status != 200:
					logger.error('Config step failed. Exiting.')
					exit(1)
			else:
				logger.debug('ADMIN user is up-to-date.')

		cls._sys_docs[ObjectId('f00000000000000000000010')] = SYS_DOC(module='user')

		# [DOC] Test if ANON user exists
		user_results = await cls.modules['user'].read(
			skip_events=[Event.PERM, Event.ON],
			env=cls._sys_env,
			query=[{'_id': 'f00000000000000000000011'}],
		)
		if not user_results.args.count:
			logger.debug('ANON user not found, creating it.')
			anon_results = await cls.modules['user'].create(
				skip_events=[Event.PERM, Event.PRE, Event.ON],
				env=cls._sys_env,
				doc=cls.compile_anon_user(),
			)
			logger.debug(f'ANON user creation results: {anon_results}')
			if anon_results.status != 200:
				logger.error('Config step failed. Exiting.')
				exit(1)
		else:
			logger.debug('ANON user found, checking it.')
			anon_doc = cls.compile_anon_user()
			anon_doc_update = {}
			for attr in cls.user_attrs.keys():
				if attr not in anon_doc or not anon_doc[attr]:
					logger.debug(f'Detected change in \'anon_doc.{attr}\' Config Attr.')
					anon_doc_update[attr] = generate_attr(attr_type=cls.user_attrs[attr])
			for module in cls.anon_privileges.keys():
				if module not in anon_doc or set(anon_doc[module]) != set(
					cls.anon_privileges[module]
				):
					logger.debug(f'Detected change in \'anon_privileges\' Config Attr.')
					anon_doc_update[f'privileges.{module}'] = cls.anon_privileges[module]
			for auth_attr in cls.user_attrs.keys():
				if (
					f'{auth_attr}_hash' not in anon_doc
					or anon_doc[f'{auth_attr}_hash'] != cls.anon_token
				):
					logger.debug(f'Detected change in \'anon_token\' Config Attr.')
					anon_doc_update[attr] = cls.anon_token
				anon_doc_update[f'{auth_attr}_hash'] = cls.anon_token
			if len(anon_doc_update.keys()):
				logger.debug(f'Attempting to update ANON user with doc: \'{anon_doc_update}\'')
				anon_results = await cls.modules['user'].update(
					skip_events=[Event.PERM, Event.PRE, Event.ON],
					env=cls._sys_env,
					query=[{'_id': ObjectId('f00000000000000000000011')}],
					doc=anon_doc_update,
				)
				logger.debug(f'ANON user update results: {anon_results}')
				if anon_results.status != 200:
					logger.error('Config step failed. Exiting.')
					exit(1)
			else:
				logger.debug('ANON user is up-to-date.')

		cls._sys_docs[ObjectId('f00000000000000000000011')] = SYS_DOC(module='user')

		logger.debug('Testing sessions collection.')
		# [DOC] Test if ANON session exists
		session_results = await cls.modules['session'].read(
			skip_events=[Event.PERM, Event.ON],
			env=cls._sys_env,
			query=[{'_id': 'f00000000000000000000012'}],
		)
		if not session_results.args.count:
			logger.debug('ANON session not found, creating it.')
			anon_results = await cls.modules['session'].create(
				skip_events=[Event.PERM, Event.PRE, Event.ON],
				env=cls._sys_env,
				doc=cls.compile_anon_session(),
			)
			logger.debug(f'ANON session creation results: {anon_results}')
			if anon_results.status != 200:
				logger.error('Config step failed. Exiting.')
				exit(1)
		cls._sys_docs[ObjectId('f00000000000000000000012')] = SYS_DOC(module='session')

		logger.debug('Testing groups collection.')
		# [DOC] Test if DEFAULT group exists
		group_results = await cls.modules['group'].read(
			skip_events=[Event.PERM, Event.ON],
			env=cls._sys_env,
			query=[{'_id': 'f00000000000000000000013'}],
		)
		if not group_results.args.count:
			logger.debug('DEFAULT group not found, creating it.')
			group_create_doc = {
				'_id': ObjectId('f00000000000000000000013'),
				'user': ObjectId('f00000000000000000000010'),
				'name': {locale: '__DEFAULT' for locale in cls.locales},
				'bio': {locale: '__DEFAULT' for locale in cls.locales},
				'privileges': cls.default_privileges,
			}
			if cls.realm:
				group_create_doc['realm'] = '__global'
			group_results = await cls.modules['group'].create(
				skip_events=[Event.PERM, Event.PRE, Event.ON],
				env=cls._sys_env,
				doc=group_create_doc,
			)
			logger.debug(f'DEFAULT group creation results: {group_results}')
			if group_results.status != 200:
				logger.error('Config step failed. Exiting.')
				exit(1)
		else:
			logger.debug('DEFAULT group found, checking it.')
			default_doc = group_results.args.docs[0]
			default_doc_update: Dict[str, Any] = {}
			for module in cls.default_privileges.keys():
				if module not in default_doc.privileges.keys() or set(
					default_doc.privileges[module]
				) != set(cls.default_privileges[module]):
					logger.debug(f'Detected change in \'default_privileges\' Config Attr.')
					default_doc_update[f'privileges.{module}'] = cls.default_privileges[module]
			if len(default_doc_update.keys()):
				logger.debug(
					f'Attempting to update DEFAULT group with doc: \'{default_doc_update}\''
				)
				default_results = await cls.modules['group'].update(
					skip_events=[Event.PERM, Event.PRE, Event.ON],
					env=cls._sys_env,
					query=[{'_id': ObjectId('f00000000000000000000013')}],
					doc=default_doc_update,
				)
				logger.debug(f'DEFAULT group update results: {default_results}')
				if anon_results.status != 200:
					logger.error('Config step failed. Exiting.')
					exit(1)
			else:
				logger.debug('DEFAULT group is up-to-date.')

		cls._sys_docs[ObjectId('f00000000000000000000013')] = SYS_DOC(module='group')

		# [DOC] Test app-specific groups
		logger.debug('Testing app-specific groups collection.')
		for group in cls.groups:
			group_results = await cls.modules['group'].read(
				skip_events=[Event.PERM, Event.ON],
				env=cls._sys_env,
				query=[{'_id': group['_id']}],
			)
			if not group_results.args.count:
				logger.debug(
					f'App-specific group with name \'{group["name"]}\' not found, creating it.'
				)
				if cls.realm:
					group['realm'] = '__global'
				group_results = await cls.modules['group'].create(
					skip_events=[Event.PERM, Event.PRE, Event.ON],
					env=cls._sys_env,
					doc=group,
				)
				logger.debug(
					f'App-specific group with name {group["name"]} creation results: {group_results}'
				)
				if group_results.status != 200:
					logger.error('Config step failed. Exiting.')
					exit(1)
			else:
				logger.debug(
					f'App-specific group with name \'{group["name"]}\' found, checking it.'
				)
				group_doc = group_results.args.docs[0]
				group_doc_update = {}
				if 'privileges' in group.keys():
					for module in group['privileges'].keys():
						if module not in group_doc.privileges.keys() or set(
							group_doc.privileges[module]
						) != set(group['privileges'][module]):
							logger.debug(
								f'Detected change in \'privileges\' Doc Arg for group with name \'{group["name"]}\'.'
							)
							group_doc_update[f'privileges.{module}'] = group['privileges'][module]
				if len(group_doc_update.keys()):
					logger.debug(
						f'Attempting to update group with name \'{group["name"]}\' with doc: \'{group_doc_update}\''
					)
					group_results = await cls.modules['group'].update(
						skip_events=[Event.PERM, Event.PRE, Event.ON],
						env=cls._sys_env,
						query=[{'_id': group['_id']}],
						doc=group_doc_update,
					)
					logger.debug(
						f'Group with name \'{group["name"]}\' update results: {group_results}'
					)
					if group_results.status != 200:
						logger.error('Config step failed. Exiting.')
						exit(1)
				else:
					logger.debug(f'Group with name \'{group["name"]}\' is up-to-date.')

			cls._sys_docs[ObjectId(group['_id'])] = SYS_DOC(module='group')

		# [DOC] Test app-specific data indexes
		logger.debug('Testing data indexes')
		for index in cls.data_indexes:
			logger.debug(f'Attempting to create data index: {index}')
			try:
				cls._sys_conn[cls.data_name][index['collection']].create_index(index['index'])
			except Exception as e:
				logger.error(f'Failed to create data index: {index}, with error: {e}')
				logger.error('Evaluate error and take action manually.')

		logger.debug(
			'Creating \'var\', \'type\', \'user\' data indexes for settings collections.'
		)
		cls._sys_conn[cls.data_name]['settings'].create_index([('var', 1)])
		cls._sys_conn[cls.data_name]['settings'].create_index([('type', 1)])
		cls._sys_conn[cls.data_name]['settings'].create_index([('user', 1)])
		logger.debug(
			'Creating \'user\', \'event\', \'subevent\' data indexes for analytics collections.'
		)
		cls._sys_conn[cls.data_name]['analytics'].create_index([('user', 1)])
		cls._sys_conn[cls.data_name]['analytics'].create_index([('event', 1)])
		cls._sys_conn[cls.data_name]['analytics'].create_index([('subevent', 1)])
		logger.debug('Creating \'__deleted\' data indexes for all collections.')
		for module in cls.modules:
			if cls.modules[module].collection:
				logger.debug(
					f'Attempting to create \'__deleted\' data index for collection: {cls.modules[module].collection}'
				)
				cls._sys_conn[cls.data_name][cls.modules[module].collection].create_index(
					[('__deleted', 1)]
				)
		if cls.realm:
			logger.debug('Creating \'realm\' data indexes for all collections.')
			for module in cls.modules:
				if module != 'realm' and cls.modules[module].collection:
					logger.debug(
						f'Attempting to create \'realm\' data index for collection: {cls.modules[module].collection}'
					)
					cls._sys_conn[cls.data_name][cls.modules[module].collection].create_index(
						[('realm', 'text')]
					)

		# [DOC] Test app-specific docs
		logger.debug('Testing docs.')
		for doc in cls.docs:
			# [TODO] Remove with final version
			if type(doc) == dict:
				logger.warning(f'Config Attr \'docs\' is using deprecated dict, Update it.')
				doc = SYS_DOC(
					module=doc['module'],
					key=doc['key'] if 'key' in doc.keys() else '_id',
					skip_args=doc['skip_args'] if 'skip_args' in doc.keys() else False,
					doc=doc['doc'],
				)

			doc_results = await cls.modules[doc.module].read(
				skip_events=[Event.PERM, Event.PRE, Event.ON, Event.ARGS],
				env=cls._sys_env,
				query=[{doc.key: doc.key_value}],  # type: ignore
			)
			if not doc_results.args.count:
				if cls.realm:
					doc.doc['realm'] = '__global'
				skip_events = [Event.PERM]
				if doc.skip_args == True:
					skip_events.append(Event.ARGS)
				doc_results = await cls.modules[doc.module].create(
					skip_events=skip_events, env=cls._sys_env, doc=doc.doc
				)
				logger.debug(
					'App-specific doc with %s \'%s\' of module \'%s\' creation results: %s',
					doc.key,
					doc.key_value,
					doc.module,
					doc_results,
				)
				if doc_results.status != 200:
					logger.error('Config step failed. Exiting.')
					exit(1)
			cls._sys_docs[ObjectId(doc_results.args.docs[0]._id)] = SYS_DOC(module=doc.module)

		# [DOC] Check for test mode
		if cls.test:
			from nawah.test import Test

			logger.debug('Running tests')
			anon_session = cls.compile_anon_session()
			anon_session['user'] = DictObj(cls.compile_anon_user())
			Test.session = DictObj(anon_session)
			Test.env = cls._sys_env
			await Test.run_test(test_name=cls.test_name)
			exit(1)

		# [DOC] Check for emulate_test mode
		if cls.emulate_test:
			cls.test = True

	@classmethod
	def compile_anon_user(cls):
		from nawah.utils import generate_attr

		anon_doc = {
			'_id': ObjectId('f00000000000000000000011'),
			'name': {cls.locale: '__ANON'},
			'groups': [],
			'privileges': cls.anon_privileges,
			'locale': cls.locale,
		}
		for attr in cls.user_attrs.keys():
			anon_doc[attr] = generate_attr(attr_type=cls.user_attrs[attr])
		for auth_attr in cls.user_attrs.keys():
			anon_doc[f'{auth_attr}_hash'] = cls.anon_token
		if cls.realm:
			anon_doc['realm'] = '__global'
		return anon_doc

	@classmethod
	def compile_anon_session(cls):
		session_doc = {
			'_id': ObjectId('f00000000000000000000012'),
			'user': ObjectId('f00000000000000000000011'),
			'host_add': '127.0.0.1',
			'user_agent': cls.anon_token,
			'timestamp': '1970-01-01T00:00:00',
			'expiry': '1970-01-01T00:00:00',
			'token': cls.anon_token,
			'token_hash': cls.anon_token,
		}
		if cls.realm:
			session_doc['realm'] = '__global'
		return session_doc



from nawah import __version__

from typing import Literal, Any

import argparse, os, logging, datetime, sys, subprocess, asyncio, traceback, shutil, urllib.request, re, tarfile, string, random, tempfile, pkgutil

logger = logging.getLogger('nawah')
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s [%(levelname)s]  %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.setLevel(logging.INFO)


def nawah_cli():
	global sys, os

	if sys.version_info.major != 3 or sys.version_info.minor < 8:
		print('Nawah framework CLI can only run with Python >= 3.8. Exiting.')
		exit(1)

	parser = argparse.ArgumentParser()
	parser.add_argument(
		'--version',
		help='Show Nawah framework version and exit',
		action='version',
		version=f'Nawah framework v{__version__}',
	)

	subparsers = parser.add_subparsers(
		title='Command', description='Nawah framework CLI command to run', dest='command'
	)

	parser_install = subparsers.add_parser(
		'install_deps', help='Install dependencies of Nawah app'
	)
	parser_install.set_defaults(func=install_deps)

	parser_launch = subparsers.add_parser('launch', help='Launch Nawah app')
	parser_launch.set_defaults(func=launch)
	parser_launch.add_argument('--env', help='Choose specific env')
	parser_launch.add_argument('--debug', help='Enable debug mode', action='store_true')
	parser_launch.add_argument(
		'--log',
		help='Enable debug mode and log all debug messages to log file',
		action='store_true',
	)
	parser_launch.add_argument('-p', '--port', help='Set custom port [default 8081]')
	parser_launch.add_argument(
		'--force-admin-check',
		help='Force ADMIN doc checked and updated, if ADMIN doc is changed',
		action='store_true',
	)
	parser_launch.add_argument(
		'--test-collections', help='Enable Test Collections Mode', action='store_true'
	)

	parser_test = subparsers.add_parser('test', help='Test Nawah app')
	parser_test.set_defaults(func=test)
	parser_test.add_argument('test_name', type=str, help='Name of the test to run')
	parser_test.add_argument('--env', help='Choose specific env')
	parser_test.add_argument(
		'--skip-flush',
		help='Skip flushing previous test data collections',
		action='store_true',
	)
	parser_test.add_argument(
		'--force',
		help='Force running all test steps even if one is failed',
		action='store_true',
	)
	parser_test.add_argument(
		'--use-env',
		help='Run tests on selected env rather than sandbox env',
		action='store_true',
	)
	parser_test.add_argument(
		'--breakpoint',
		help='Create debugger breakpoint upon failure of test',
		action='store_true',
	)
	parser_test.add_argument('--debug', help='Enable debug mode', action='store_true')

	parser_packages = subparsers.add_parser('packages', help='Manage Nawah app packages')
	parser_packages.set_defaults(func=lambda _: None)
	packages_subparser = parser_packages.add_subparsers(
		title='Packages Command',
		description='Packages command to run',
		dest='packages_command',
	)

	parser_packages_add = packages_subparser.add_parser(
		'add', help='Add package to Nawah app'
	)
	parser_packages_add.set_defaults(func=packages_add)
	parser_packages_add.add_argument(
		'package', help='Package identifier/URI/path to add to Nawah app'
	)
	parser_packages_add.add_argument(
		'--source',
		help='Package source [default nawah]',
		choices=['nawah', 'local'],
		default='nawah',
	)

	parser_packages_rm = packages_subparser.add_parser(
		'rm', help='Remove package from Nawah app'
	)
	parser_packages_rm.set_defaults(func=packages_rm)
	parser_packages_rm.add_argument(
		'package_name', help='Package name to remove from Nawah app'
	)

	parser_packages_audit = packages_subparser.add_parser(
		'audit', help='Audit packages in Nawah app'
	)
	parser_packages_audit.set_defaults(func=packages_audit)

	parser_ref = subparsers.add_parser('generate_ref', help='Generate Nawah app reference')
	parser_ref.set_defaults(func=generate_ref)
	parser_ref.add_argument('--debug', help='Enable debug mode', action='store_true')

	parser_ref = subparsers.add_parser('generate_models', help='Generate Nawah app models')
	parser_ref.set_defaults(func=generate_models)
	parser_ref.add_argument('format', help='Format of models', choices=['js', 'ts'])
	parser_ref.add_argument('--debug', help='Enable debug mode', action='store_true')

	args = parser.parse_args()

	if args.command:
		if args.command == 'packages' and not args.packages_command:
			parser_packages.print_help()
		else:
			args.func(args)
	else:
		parser.print_help()


def install_deps(args: argparse.Namespace):
	global sys, os, subprocess
	# [DOC] Change logging level to debug
	logger.setLevel(logging.DEBUG)
	logger.debug('Beginning to install dependencies')
	# [DOC] Create standard call command list
	pip_command = [sys.executable, '-m', 'pip', 'install', '--user', '-r']

	dirs = [
		d
		for d in os.listdir(os.path.join('.', 'packages'))
		if os.path.isdir(os.path.join('.', 'packages', d))
	]
	# [DOC] Iterate over packages to find requirements.txt files
	for package in dirs:
		logger.debug(f'Checking package \'{package}\' for \'requirements.txt\' file.')
		if os.path.exists(os.path.join('.', 'packages', package, 'requirements.txt')):
			logger.debug(
				'File \'requirements.txt\' found! Attempting to install package dependencies.'
			)
			pip_call = subprocess.call(
				pip_command + [os.path.join('.', 'packages', package, 'requirements.txt')]
			)
			if pip_call != 0:
				logger.error('Last \'pip\' call failed. Check console for more details. Exiting.')
				exit(1)


def launch(
	args: argparse.Namespace,
	custom_launch: Literal['test', 'generate_ref', 'generate_models'] = None,
):
	global os, asyncio
	global handler

	# [DOC] Update Config with Nawah CLI args
	from nawah.config import Config, process_config

	Config._nawah_version = __version__
	if custom_launch not in ['generate_ref', 'generate_models']:
		Config.env = args.env
	if not custom_launch:
		Config.test_collections = args.test_collections
		Config.force_admin_check = args.force_admin_check

	# [DOC] Check for debug CLI Arg
	if args.debug:
		Config.debug = True
		logger.setLevel(logging.DEBUG)
	# [DOC] Check for log CLI Arg
	if not custom_launch and args.log:
		logger.removeHandler(handler)
		if not os.path.exists(os.path.join('.', 'logs')):
			os.makedirs(os.path.join('.', 'logs'))
		handler = logging.FileHandler(
			filename=os.path.join(
				'.',
				'logs',
				f'{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.log',
			)
		)
		handler.setFormatter(formatter)
		logger.addHandler(handler)
		logger.setLevel(logging.DEBUG)

	from nawah.app import run_app

	try:
		try:
			sys.path.append('.')
			nawah_app = __import__('nawah_app')
			app_config = nawah_app.config
			Config._app_packages = nawah_app.packages
		except ModuleNotFoundError:
			logger.error(f'No \'nawah_app.py\' file found in CWD. Exiting.')
			exit(1)
		except AttributeError:
			logger.error(
				f'File \'nawah_app.py\' was found but it doesn\'t have \'config\' method. Exiting.'
			)
			exit(1)
		logger.info(
			f'Found app \'{app_config.name} (v{app_config.version})\'. Attempting to load App Config.'
		)
		Config._app_name = app_config.name
		Config._app_version = app_config.version
		Config._app_default_package = app_config.default_package

		# [DOC] Check all required values of App Config are present
		if not Config._app_name or not Config._app_version or not Config._app_default_package:
			logger.error(
				f'App Config are missing at least one of the following \'name, version, default_package\'. Exiting.'
			)
			exit(1)

		Config._app_path = os.path.realpath('.')
		# [DOC] Read app_config and update Config accordingly
		# [DOC] Check envs, env
		if custom_launch not in ['generate_ref', 'generate_models'] and app_config.envs:
			if not args.env and not app_config.env:
				logger.error(
					'App Config Attr \'envs\' found, but no \'env\' App Config Attr, or CLI Attr were defined.'
				)
				exit(1)
			if args.env:
				if args.env in app_config.envs.keys():
					logger.info(f'Setting \'env\' Config Attr to \'env\' CLI Arg value \'{args.env}\'')
				else:
					logger.error(
						f'Found value \'{args.env}\' for \'env\' CLI Arg, but not defined in \'envs\' App Config Attr. Exiting.'
					)
					exit(1)
			else:
				if app_config.env in app_config.envs.keys():
					logger.info(
						f'Setting \'env\' Config Attr to \'env\' App Config Attr value \'{app_config.env}\''
					)
					Config.env = app_config.env
				elif app_config.env.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'env\' App Config Attr. Attempting to process it.'
					)
					env_env_var = app_config.env.replace('$__env.', '')
					env = os.getenv(env_env_var)
					if env:
						logger.info(
							f'Setting \'env\' Config Attr to Env Variable \'{env_env_var}\' value \'{env}\'.'
						)
						Config.env = env
					else:
						logger.error(f'No value found for Env Variable \'{env_env_var}\'. Exiting.')
						exit(1)
				else:
					logger.error(
						f'Found value \'{args.env}\' for \'env\' CLI Arg, but not defined in \'envs\' App Config Attr. Exiting.'
					)
					exit(1)
			logger.info(
				f'Beginning to extract Config Attrs defined in selected \'env\', \'{Config.env}\', to App Config Attrs.'
			)
			for config_attr in dir(app_config.envs[Config.env]):
				if (
					config_attr.startswith('__')
					or getattr(app_config.envs[Config.env], config_attr) == None
				):
					continue
				logger.info(f'Extracting \'{config_attr}\' Config Attr to App Config Attr')
				setattr(
					app_config,
					config_attr,
					getattr(app_config.envs[Config.env], config_attr),
				)
				setattr(
					Config,
					config_attr,
					getattr(app_config.envs[Config.env], config_attr),
				)
		# [DOC] Check port Config Attr
		if not custom_launch and app_config.port:
			if args.port:
				logger.info(
					f'Ignoring \'port\' App Config Attr in favour of \'port\' CLI Arg with value \'{args.port}\'.'
				)
				try:
					Config.port = int(args.port)
				except:
					logger.error(f'Port should be in integer format. Exiting.')
					exit(1)
			else:
				logger.info('Found \'port\' App Config Attr. Attempting to process it.')
				if type(app_config.port) == int:
					Config.port = app_config.port
					logger.info(f'Setting \'port\' Config Attr to \'{Config.port}\'.')
				elif type(app_config.port) == str and app_config.port.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'port\' App Config Attr. Attempting to process it.'
					)
					port_env_var = app_config.port.replace('$__env.', '')
					port = os.getenv(port_env_var)
					if port:
						logger.info(
							f'Setting \'port\' Config Attr to Env Variable \'{port_env_var}\' value \'{port}\'.'
						)
						try:
							Config.port = int(port)
						except:
							logger.error(
								f'Env Variable \'{port_env_var}\' value \'{port}\' can\'t be converted to integer. Exiting.'
							)
							exit(1)
					else:
						logger.error(f'No value found for Env Variable \'{port_env_var}\'. Exiting.')
						exit(1)
				else:
					logger.error(
						f'Invalid value type for \'port\' Config Attr with value \'{app_config.port}\'. Exiting.'
					)
					exit(1)
		# [DOC] Check debug Config Attr
		if app_config.debug:
			if args.debug:
				logger.info(
					f'Ignoring \'debug\' App Config Attr in favour of \'debug\' CLI Arg with value \'{args.debug}\'.'
				)
				Config.debug = args.debug
			else:
				logger.info('Found \'debug\' App Config Attr. Attempting to process it.')
				if type(app_config.debug) == bool:
					Config.debug = app_config.debug
					logger.info(f'Setting \'debug\' Config Attr to \'{Config.debug}\'.')
				elif type(app_config.debug) == str and app_config.debug.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'debug\' App Config Attr. Attempting to process it.'
					)
					debug_env_var = app_config.debug.replace('$__env.', '')
					debug = os.getenv(debug_env_var)
					if debug:
						logger.info(
							f'Setting \'debug\' Config Attr to Env Variable \'{debug_env_var}\' as \'True\'.'
						)
						Config.debug = True
					else:
						logger.info(
							f'No value found for Env Variable \'{debug_env_var}\'. Setting \'debug\' to \'False\'.'
						)
						Config.debug = False
				else:
					logger.error(
						f'Invalid value type for \'debug\' Config Attr with value \'{app_config.debug}\'. Exiting.'
					)
					exit(1)
				if Config.debug:
					logger.setLevel(logging.DEBUG)
		# [DOC] Check force_admin_check Config Attr
		if not custom_launch and app_config.force_admin_check:
			if args.force_admin_check:
				logger.info(
					f'Ignoring \'force_admin_check\' App Config Attr in favour of \'force_admin_check\' CLI Arg with value \'{args.force_admin_check}\'.'
				)
				Config.force_admin_check = True
			else:
				logger.info(
					'Found \'force_admin_check\' App Config Attr. Attempting to process it.'
				)
				if type(app_config.force_admin_check) == bool:
					Config.force_admin_check = app_config.force_admin_check
					logger.info(
						f'Setting \'force_admin_check\' Config Attr to \'{Config.force_admin_check}\'.'
					)
				elif type(
					app_config.force_admin_check
				) == str and app_config.force_admin_check.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'force_admin_check\' App Config Attr. Attempting to process it.'
					)
					check_env_var = app_config.force_admin_check.replace('$__env.', '')
					check = os.getenv(check_env_var)
					if check:
						logger.info(
							f'Setting \'force_admin_check\' Config Attr to Env Variable \'{check_env_var}\' as \'True\'.'
						)
						Config.force_admin_check = True
					else:
						logger.info(
							f'No value found for Env Variable \'{check_env_var}\'. Setting \'force_admin_check\' to \'False\'.'
						)
						Config.force_admin_check = False
				else:
					logger.error(
						f'Invalid value type for \'force_admin_check\' Config Attr with value \'{app_config.force_admin_check}\'. Exiting.'
					)
					exit(1)
		# [TODO] Implement realm APP Config Attr checks
		# [DOC] Process other app config attrs as PACKAGE_CONFIG
		process_config(config=app_config)
	except Exception:
		logger.error(
			'An unexpected exception happened while attempting to process Nawah app. Exception details:'
		)
		logger.error(traceback.format_exc())
		logger.error('Exiting.')
		exit(1)

	asyncio.run(run_app())


def test(args: argparse.Namespace):
	# [DOC] Update Config with Nawah framework CLI args
	from nawah.config import Config

	Config.test = True
	Config.test_name = args.test_name
	Config.test_skip_flush = args.skip_flush
	Config.test_force = args.force
	Config.test_env = args.use_env
	Config.test_breakpoint = args.breakpoint
	launch(args=args, custom_launch='test')


def packages_add(args: argparse.Namespace):
	logger.info('Checking packages conflicts.')
	package_name = args.package
	package_path = os.path.realpath(os.path.join('.', 'packages', package_name))

	if os.path.exists(package_path):
		logger.error(f'Package \'{package_name}\' already exists in app. Exiting.')
		exit(1)

	api_level = '.'.join(__version__.split('.')[:2])

	if args.source == 'nawah':
		logger.info(f'Attempting to add package from Nawah source.')
		package_root = f'{package_name}-APIv{api_level}'
		package_url = (
			f'https://github.com/nawah-io/{package_name}/archive/APIv{api_level}.tar.gz'
		)
		logger.info(f'Attempting to download package from: {package_url}')
		# [REF] https://stackoverflow.com/a/7244263/2393762
		package_archive, _ = urllib.request.urlretrieve(package_url)
		logger.info('Package archive downloaded successfully!')

	elif args.source == 'local':
		logger.info(f'Attempting to add package from local source: \'{args.package}\'.')
		try:
			source_path = os.path.realpath(args.package)
			package_name = package_root = os.path.basename(source_path)
			sys.path.insert(0, source_path)
			package = __import__('__init__')
		except:
			logger.error('Failed to load package. Exiting.')
			exit(1)

		logger.info(f'Attempting to check package API Level matching \'{api_level}\'.')
		if package.config.api_level != api_level:
			logger.error(
				f'Package is using incompatible API Level \'{package.config.api_level}\'. Exiting.'
			)
			exit(1)

		logger.info('Attempting to archive package from local source.')
		temp_package_archive = tempfile.NamedTemporaryFile(mode='w')
		# [REF] https://stackoverflow.com/a/17081026
		# [REF] https://stackoverflow.com/a/16000963
		with tarfile.open(temp_package_archive.name, 'w:gz') as archive:
			archive.add(
				source_path,
				arcname=package_name,
				filter=lambda member: None if '.git/' in member.name else member,
			)
		logger.info('Package archive created successfully!')
		package_archive = temp_package_archive.name

	def archive_members(
		*, archive: tarfile.TarFile, root_path: str, search_path: str = None
	):
		l = len(f'{root_path}/')
		for member in archive.getmembers():
			if member.path.startswith(f'{root_path}/{search_path or ""}'):
				member.path = member.path[l:]
				yield member

	logger.info('Attempting to extract package archive to Nawah app \'packages\'.')
	with tarfile.open(name=package_archive, mode='r:gz') as archive:
		archive.extractall(
			path=package_path,
			members=archive_members(archive=archive, root_path=package_root),
		)
	logger.info('Package archive extracted successfully!')

	logger.info('Attempting to update \'nawah_packages.py\'.')

	sys.path.insert(0, os.path.realpath('.'))
	sys.path.insert(0, package_path)
	package = __import__('__init__')
	import nawah_packages

	packages = nawah_packages.packages
	packages[package_name] = package.config.version
	with open(os.path.realpath(os.path.join('.', 'nawah_packages.py')), 'w') as f:
		packages_string = ''
		for package in packages.keys():
			packages_string += f'    \'{package}\': \'{packages[package]}\',\n'
		f.write(f'packages = {{\n{packages_string}}}\n')
	logger.info('Successfully updated \'nawah_packages.py\'.')
	logger.info(
		'Remember to check package docs for any \'vars\' you are required to add to your app.'
	)


def packages_rm(args: argparse.Namespace):
	logger.warning(
		'Due to compatibility issues you will be required to manually remove the package from \'packages\' and \'nawah_packages.py\'. Exiting.'
	)
	exit(1)


def packages_audit(args: argparse.Namespace):
	logger.info(
		'Attempting to audit packages in \'nawah_pacakges.py\' against packages added to \'packages\' folder.'
	)

	app_path = os.path.realpath('.')
	packages_path = os.path.realpath(os.path.join('.', 'packages'))
	sys.path.insert(0, app_path)
	sys.path.append(packages_path)
	import nawah_app, nawah_packages

	req_packages = nawah_packages.packages
	logger.info('Packages in \'nawah_pacakges.py\' are:')
	for package in req_packages.keys():
		logger.info(f'- {package}: {req_packages[package]}')

	logger.info('Attempting to check packages added in \'packages\'.')
	added_packages = {}
	for _, pkgname, ispkg in pkgutil.iter_modules([packages_path]):
		if not ispkg:
			continue
		logger.info(f'- Importing package: {pkgname}')
		package = __import__(pkgname, fromlist='*')
		added_packages[pkgname] = package.config.version
		logger.info(f'- Imported package with version: {package.config.version}')

	logger.info('Attempting to audit missing, extra, version-mismatching packages.')

	missing_packages = [
		package for package in req_packages if package not in added_packages.keys()
	]
	if missing_packages:
		logger.info(
			'Following packages are required by your app but not added to \'packages\':'
		)
		for package in missing_packages:
			logger.info(f'- {package}')
	else:
		logger.info('Great! Your app is not missing any of the required package.')

	extra_packages = [
		package
		for package in added_packages
		if package not in req_packages.keys() and package != nawah_app.config.default_package
	]
	if extra_packages:
		logger.info(
			'Following packages are added to \'packages\' but not required by your app:'
		)
		for package in extra_packages:
			logger.info(f'- {package}')
	else:
		logger.info('Great! Your app is not having any extra package.')

	mismatch_packages = [
		package
		for package in req_packages
		if package in added_packages.keys()
		and added_packages[package] != req_packages[package]
	]
	if mismatch_packages:
		logger.warning(
			'Following packages are having mismatch between required and added version:'
		)
		for package in mismatch_packages:
			logger.info(
				f'- {package} requires version \'{req_packages[package]}\' but \'{added_packages[package]}\' is added'
			)
	else:
		logger.info('Great! Your app is not having any version-mismatching package.')


def generate_ref(args: argparse.Namespace):
	# [DOC] Update Config with Nawah framework CLI args
	from nawah.config import Config

	Config.generate_ref = True
	launch(args=args, custom_launch='generate_ref')


def generate_models(args: argparse.Namespace):
	# [DOC] Update Config with Nawah framework CLI args
	from nawah.config import Config

	Config.generate_models = True
	launch(args=args, custom_launch='generate_models')
